import platform
import zipfile
from pathlib import Path
from urllib.parse import urlparse
from urllib.request import url2pathname
from zipfile import ZipFile

from compliance_checker.protocols import netcdf


def _fix_windows_slashes(zarr_url):
    if platform.system() == "Windows":
        zarr_url = zarr_url.replace("///", "//")
    return zarr_url


def is_zarr(url):
    """Check if a URL is a zarr dataset."""
    url = str(url)

    if netcdf.is_netcdf(url):
        return False

    if ".zarr" in url:
        return True

    if urlparse(url).scheme in ("https", "s3", "file"):
        return True

    if zipfile.is_zipfile(url):
        if ".zmetadata" in ZipFile(url).namelist():
            return True

    if Path(url).is_dir():
        if (Path(url) / ".zmetadata").exists():
            return True

    return False


def as_zarr(url):
    """
    Transform pointers to zarr datasets to valid nczarr urls, as described in
    https://www.unidata.ucar.edu/blogs/developer/entry/overview-of-zarr-support-in\n
    url: str or Path to valid zarr dataset\n
    Distinct from is_cdl etc in that it will return the appropriate URI \n\n

    Not tested on Windows paths at the moment, as NCZarr is not supported in Windows\n

    A valid Zarr dataset could be provided in any of the following forms:\n
    "http://s3.amazonaws.com/bucket/dataset.zarr"\n
    "http://s3.amazonaws.com/bucket/dataset.zarr"#mode=nczarr,s3\n
    "/home/path/to/dataset.zarr"\n
    Path('/home/path/to/dataset.zarr')\n
    "file:///home/path/to/dataset.zarr"\n
    "file:///home/path/to/dataset.randomExt#mode=nczarr,file"
    "file:///home/path/to/dataset.zarr#mode=nczarr,zip"
    """

    pr = urlparse(str(url))

    if "mode=nczarr" in pr.fragment:
        if pr.netloc or pr.scheme == "file":
            url = _fix_windows_slashes(url)
        return url

    zarr_url = Path(
        url2pathname(pr.path),
    ).resolve()  # url2pathname necessary to avoid urlparse bug in windows

    if pr.netloc:
        mode = "s3"
    elif zipfile.is_zipfile(zarr_url):
        mode = "zip"
    elif zarr_url.is_dir():
        mode = "file"
    else:
        raise ValueError(
            f"Could not identify {url},\nif #mode=nczarr,zarr, please pass this explicitly\nValid url options are described here\nhttps://www.unidata.ucar.edu/blogs/developer/entry/overview-of-zarr-support-in",
        )

    url_base = url if mode == "s3" else zarr_url.as_uri()

    zarr_url = f"{url_base}#mode=nczarr,{mode}"
    zarr_url = _fix_windows_slashes(zarr_url)
    return zarr_url



from collections.abc import Generator
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from typing import BinaryIO

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"


@contextmanager
def tempnc(data: BinaryIO) -> Generator[str, None, None]:
    """
    Create a temporary in-memory NetCDF file using a NamedTemporaryFile.
    Close the file automatically after scope is exited.

    Type aliasing and tempfile creation credit to @ocefpaf
    https://github.com/ioos/compliance-checker/pull/799#discussion_r411420587

    Parameters
    ----------
    data (bytes): raw bytes to store in the NamedTemporaryFile

    Returns
    -------
    context-managed generator
    """
    tmp = None
    try:
        tmp = NamedTemporaryFile(suffix=".nc", prefix="compliance-checker_")
        tmp.write(data)
        tmp.flush()
        yield tmp.name
    finally:
        if tmp is not None:
            tmp.close()



def is_tabledap(url):
    """
    Identify a dataset as an ERDDAP TableDAP dataset.

    Parameters
    ----------
    url (str) : URL to dataset

    Returns
    -------
    bool
    """

    return "tabledap" in url



import subprocess
from importlib.resources import files


def get_filename(path):
    """
    Returns the path to a valid dataset
    """
    filename = files("compliance_checker") / path
    nc_path = filename.with_suffix(".nc")
    if not nc_path.exists():
        generate_dataset(filename, nc_path)
    return nc_path


def generate_dataset(cdl_path, nc_path):
    subprocess.call(["ncgen", "-4", "-o", nc_path, cdl_path])


STATIC_FILES = {
    # TODO: add defaulltdict implementation for default files, etc
    "bad": get_filename("tests/data/non-comp/bad.cdl"),
    "badname": get_filename("tests/data/non-comp/badname.netcdf"),
    "bad-instance": get_filename("tests/data/bad-instance.cdl"),
    "bad-rhgrid": get_filename("tests/data/non-comp/bad-rhgrid.cdl"),
    "bad-trajectory": get_filename("tests/data/bad-trajectory.cdl"),
    "bad_cell_measure1": get_filename("tests/data/bad_cell_measure1.cdl"),
    "bad_cell_measure2": get_filename("tests/data/bad_cell_measure2.cdl"),
    "bad_cf_role": get_filename("tests/data/bad_cf_role.cdl"),
    "bad_data_type": get_filename("tests/data/bad_data_type.cdl"),
    "bad_missing_data": get_filename("tests/data/bad_missing_data.cdl"),
    "bad_reference": get_filename("tests/data/bad_reference.cdl"),
    "bad_region": get_filename("tests/data/bad_region.cdl"),
    "bad_units": get_filename("tests/data/bad_units.cdl"),
    "bad2dim": get_filename("tests/data/non-comp/bad2dim.cdl"),
    "bounds_bad_order": get_filename("tests/data/non-comp/bounds_bad_order.cdl"),
    "bounds_bad_num_coords": get_filename(
        "tests/data/non-comp/bounds_bad_num_coords.cdl",
    ),
    "cell_measure": get_filename("tests/data/cell_measure.cdl"),
    "cf_example_cell_measures": get_filename(
        "tests/data/examples/cf_example_cell_measures.cdl",
    ),
    "chap2": get_filename("tests/data/chap2.cdl"),
    "climatology": get_filename("tests/data/climatology.cdl"),
    "cont_ragged": get_filename("tests/data/cont_ragged.cdl"),
    "conv_multi": get_filename("tests/data/conv_multi.cdl"),
    "conv_bad": get_filename("tests/data/conv_bad.cdl"),
    "coordinates_and_metadata": get_filename("tests/data/coordinates_and_metadata.cdl"),
    "coordinate_types": get_filename("tests/data/coordinate_types.cdl"),
    "dimensionless": get_filename("tests/data/dimensionless.cdl"),
    "duplicate_axis": get_filename("tests/data/duplicate_axis.cdl"),
    "dimension_order": get_filename("tests/data/dimension_order.cdl"),
    "example-grid": get_filename("tests/data/example-grid.cdl"),
    "featureType": get_filename("tests/data/example-grid.cdl"),
    "forecast_reference": get_filename("tests/data/forecast_reference.cdl"),
    "fvcom": get_filename("tests/data/examples/fvcom.cdl"),
    "ghrsst": get_filename(
        "tests/data/20160919092000-ABOM-L3S_GHRSST-SSTfnd-AVHRR_D-1d_dn_truncate.cdl",
    ),
    "glcfs": get_filename("tests/data/examples/glcfs.cdl"),
    "grid-boundaries": get_filename("tests/data/grid-boundaries.cdl"),
    "grid_mapping_coordinates": get_filename("tests/data/grid_mapping_coordinates.cdl"),
    "hycom_global": get_filename("tests/data/examples/hycom_global.cdl"),
    "h_point": get_filename("tests/data/appendix_h/point.cdl"),
    "h_timeseries-incomplete": get_filename(
        "tests/data/appendix_h/timeseries-incomplete.cdl",
    ),
    "h_timeseries-orthogonal": get_filename(
        "tests/data/appendix_h/timeseries-orthogonal.cdl",
    ),
    "h_timeseries-single": get_filename("tests/data/appendix_h/timeseries-single.cdl"),
    "illegal-vertical": get_filename("tests/data/illegal-vertical.cdl"),
    "illegal-aux-coords": get_filename("tests/data/illegal-aux-coords.cdl"),
    "index_ragged": get_filename("tests/data/index_ragged.cdl"),
    "index_ragged2": get_filename("tests/data/index_ragged2.cdl"),
    "indexed_ragged_domain": get_filename("tests/data/indexed_ragged_domain.cdl"),
    "ints64": get_filename("tests/data/ints64.cdl"),
    "ioos_gold_1_1": get_filename("tests/data/ioos_1_1.cdl"),
    "kibesillah": get_filename("tests/data/examples/kibesillah.cdl"),
    "l01-met": get_filename("tests/data/examples/l01-met.cdl"),
    "lateral_search_example": get_filename(
        "tests/data/examples/lateral_search_example",
    ),
    "line_geometry": get_filename("tests/data/line_geometry.cdl"),
    "mapping": get_filename("tests/data/mapping.cdl"),
    "multi-dim-coordinates": get_filename("tests/data/multi-dim-coordinates.cdl"),
    "multi-timeseries-orthogonal": get_filename(
        "tests/data/multi-timeseries-orthogonal.cdl",
    ),
    "multi-timeseries-incomplete": get_filename(
        "tests/data/multi-timeseries-incomplete.cdl",
    ),
    "ncei_gold_point_1": get_filename("tests/data/ncei_gold_point_1.cdl"),
    "ncei_gold_point_2": get_filename("tests/data/ncei_gold_point_2.cdl"),
    "NCEI_profile_template_v2_0": get_filename(
        "tests/data/NCEI_profile_template_v2.0_2016-09-22_181835.151325.cdl",
    ),
    "ocos": get_filename("tests/data/examples/ocos.cdl"),
    "ooi_glider": get_filename("tests/data/examples/ooi_glider.cdl"),
    "point": get_filename("tests/data/point.cdl"),
    "polygon_geometry": get_filename("tests/data/polygon_geometry.cdl"),
    "profile-orthogonal": get_filename("tests/data/profile-orthogonal.cdl"),
    "profile-incomplete": get_filename("tests/data/profile-incomplete.cdl"),
    "pr_inundation": get_filename("tests/data/examples/pr_inundation.cdl"),
    "reduced_horizontal_grid": get_filename("tests/data/reduced_horizontal_grid.cdl"),
    "rotated_pole_grid": get_filename("tests/data/rotated_pole_grid.cdl"),
    "rhgrid": get_filename("tests/data/rhgrid.cdl"),
    "rutgers": get_filename("tests/data/ru07-20130824T170228_rt0.cdl"),
    "scalar_coordinate_variable": get_filename(
        "tests/data/scalar_coordinate_variable.cdl",
    ),
    "self-referencing-var": get_filename("tests/data/self-referencing-var.cdl"),
    "self_referencing": get_filename("tests/data/non-comp/self_referencing.cdl"),
    "sldmb_43093_agg": get_filename("tests/data/examples/sldmb_43093_agg.cdl"),
    "string": get_filename("tests/data/string_type_variable.cdl"),
    "swan": get_filename("tests/data/examples/swan.cdl"),
    "sp041": get_filename("tests/data/examples/sp041.cdl"),
    "taxonomy_example": get_filename("tests/data/taxonomy_example.cdl"),
    "timeseries": get_filename("tests/data/timeseries.cdl"),
    "timeseries-profile-single-station": get_filename(
        "tests/data/timeseries-profile-single-station.cdl",
    ),
    "timeseries-profile-multi-station": get_filename(
        "tests/data/timeseries-profile-multi-station.cdl",
    ),
    "timeseries-profile-single-ortho-time": get_filename(
        "tests/data/timeseries-profile-single-ortho-time.cdl",
    ),
    "timeseries-profile-multi-ortho-time": get_filename(
        "tests/data/timeseries-profile-multi-ortho-time.cdl",
    ),
    "timeseries-profile-ortho-depth": get_filename(
        "tests/data/timeseries-profile-ortho-depth.cdl",
    ),
    "timeseries-profile-incomplete": get_filename(
        "tests/data/timeseries-profile-incomplete.cdl",
    ),
    "time_units": get_filename("tests/data/non-comp/time_units.cdl"),
    "trajectory-complete": get_filename("tests/data/trajectory-complete.cdl"),
    "trajectory-implied": get_filename("tests/data/trajectory-implied.cdl"),
    "trajectory-profile-orthogonal": get_filename(
        "tests/data/trajectory-profile-orthogonal.cdl",
    ),
    "trajectory-profile-incomplete": get_filename(
        "tests/data/trajectory-profile-incomplete.cdl",
    ),
    "trajectory": get_filename("tests/data/trajectory.cdl"),
    "trajectory-single": get_filename("tests/data/trajectory-single.cdl"),
    "units_check": get_filename("tests/data/units_check.cdl"),
    "usgs_dem_saipan": get_filename("tests/data/examples/usgs_dem_saipan.cdl"),
    "valid_coordinates": get_filename("tests/data/valid_coordinates.cdl"),
    "vertical_coords": get_filename("tests/data/vertical_coords.cdl"),
    "ww3": get_filename("tests/data/examples/ww3.cdl"),
    "1d_bound_bad": get_filename("tests/data/non-comp/1d_bound_bad.cdl"),
    "2dim": get_filename("tests/data/2dim-grid.cdl"),
    "2d-regular-grid": get_filename("tests/data/2d-regular-grid.cdl"),
    "2d-static-grid": get_filename("tests/data/2d-static-grid.cdl"),
    "3d-regular-grid": get_filename("tests/data/3d-regular-grid.cdl"),
    "3d-static-grid": get_filename("tests/data/3d-static-grid.cdl"),
    "3mf07": get_filename("tests/data/examples/3mf07.cdl"),
}



import os
import subprocess
from importlib.resources import files
from itertools import chain

import pytest
from netCDF4 import Dataset

from compliance_checker.cf import util
from compliance_checker.suite import CheckSuite

datadir = files("compliance_checker").joinpath("tests/data").resolve()


def glob_down(pth, suffix, lvls):
    """globs down up to (lvls: int) levels of subfolders\n
    suffix in the form ".ipynb"\n
    pth: Path"""
    return list(chain(*[pth.glob(f'*{"/*"*lvl}{suffix}') for lvl in range(lvls)]))


def generate_dataset(cdl_path, nc_path):
    subprocess.call(["ncgen", "-4", "-o", str(nc_path), str(cdl_path)])


def static_files(cdl_stem):
    """
    Returns the Path to a valid nc dataset\n
    replaces the old STATIC_FILES dict
    """
    datadir = files("compliance_checker").joinpath("tests/data").resolve()
    assert datadir.exists(), f"{datadir} not found"

    cdl_paths = glob_down(datadir, f"{cdl_stem}.cdl", 3)
    assert (
        len(cdl_paths) > 0
    ), f"No file named {cdl_stem}.cdl found in {datadir} or its subfolders"
    assert (
        len(cdl_paths) == 1
    ), f"Multiple candidates found with the name {cdl_stem}.cdl:\n{cdl_paths}\nPlease reconcile naming conflict"
    cdl_path = cdl_paths[0]  # PurePath object

    nc_path = cdl_path.parent / f"{cdl_path.stem}.nc"
    if not nc_path.exists():
        generate_dataset(cdl_path, nc_path)
        assert (
            nc_path.exists()
        ), f"ncgen CLI utility failed to produce {nc_path} from {cdl_path}"
    return str(nc_path)


# ---------Fixtures-----------

# class scope:


@pytest.fixture
def cs(scope="class"):
    """
    Initialize the dataset
    """
    cs = CheckSuite()
    cs.load_all_available_checkers()
    return cs


@pytest.fixture
def std_names(scope="class"):
    """get current std names table version (it changes)"""
    _std_names = util.StandardNameTable()
    return _std_names


# func scope:


@pytest.fixture
def loaded_dataset(request):
    """
    Return a loaded NC Dataset for the given path\n
    nc_dataset_path parameterized for each test
    """
    nc_dataset_path = static_files(request.param)
    nc = Dataset(nc_dataset_path, "r")
    yield nc
    nc.close()


@pytest.fixture
def new_nc_file(tmpdir):
    """
    Make a new temporary netCDF file for the scope of the test
    """
    nc_file_path = os.path.join(tmpdir, "example.nc")
    if os.path.exists(nc_file_path):
        raise OSError(f"File Exists: {nc_file_path}")
    nc = Dataset(nc_file_path, "w")
    # no need for cleanup, built-in tmpdir fixture will handle it
    return nc


@pytest.fixture
def tmp_txt_file(tmpdir):
    file_path = os.path.join(tmpdir, "output.txt")
    if os.path.exists(file_path):
        raise OSError(f"File Exists: {file_path}")

    return file_path


@pytest.fixture
def checksuite_setup():
    """For test_cli"""
    CheckSuite.checkers.clear()
    CheckSuite.load_all_available_checkers()



#!/usr/bin/env python
"""
compliance_checker/protocols/netcdf.py

Functions to assist in determining if the URL points to a netCDF file
"""

import warnings

import requests


def is_netcdf(url):
    """
    Returns True if the URL points to a valid local netCDF file

    :param str url: Location of file on the file system
    """
    # Try an obvious exclusion of remote resources
    url = str(url)
    if url.startswith("http"):
        return False

    # If it's a known extension, give it a shot
    if url.endswith("nc"):
        return True

    try:
        # Brute force
        with open(url, "rb") as f:
            magic_number = f.read(4)
            if len(magic_number) < 4:
                return False
            if is_classic_netcdf(magic_number):
                return True
            elif is_hdf5(magic_number):
                return True
    except PermissionError:
        # open will fail for both a directory or a local url, either of which may be pointing to a Zarr dataset
        # directory
        return False
    except OSError:
        # local file url
        return False

    return False


def is_classic_netcdf(file_buffer):
    """
    Returns True if the contents of the byte array matches the magic number in
    netCDF files

    :param str file_buffer: Byte-array of the first 4 bytes of a file
    """
    # CDF.
    if file_buffer == b"\x43\x44\x46\x01":
        return True
    return False


def is_hdf5(file_buffer):
    """
    Returns True if the contents of the byte array matches the magic number in
    HDF5 files

    :param str file_buffer: Byte-array of the first 4 bytes of a file
    """
    # .HDF
    if file_buffer == b"\x89\x48\x44\x46":
        return True
    return False


def is_remote_netcdf(ds_str):
    """
    Check a remote path points to a NetCDF resource.

    Parameters
    ----------
    ds_str (str): remote path to a dataset

    Returns
    -------
    bool
    """

    # Some datasets do not support HEAD requests!  The vast majority will,
    # however, support GET requests
    try:
        head_req = requests.head(ds_str, allow_redirects=True, timeout=10)
        head_req.raise_for_status()
    except requests.exceptions.RequestException as e:
        warnings.warn(
            f"Received exception when making HEAD request to {ds_str}: {e}",
            stacklevel=2,
        )
        content_type = None
    else:
        content_type = head_req.headers.get("content-type")

    if content_type is None:
        return False

    # if the Content-Type header returned was "application/x-netcdf",
    # or a netCDF file (not OPeNDAP) we can open this into a Dataset
    # Add support for application/x-netcdf;ver=4
    return content_type.split(";")[0] == "application/x-netcdf"



#!/usr/bin/env python
"""Tests for base compliance checker class"""

import os

from netCDF4 import Dataset

from compliance_checker import base


class TestBase:
    """
    Tests functionality of the base compliance checker class
    """

    def setup_method(self):
        self.acdd = base.BaseCheck()
        self.ds = Dataset(filename=os.devnull, mode="w", diskless=True)

    def teardown_method(self):
        self.ds.close()

    def test_attr_presence(self):
        # attribute not present, should fail
        priority = base.BaseCheck.MEDIUM
        rv1, rv2, rv3, rv4 = [], [], [], []
        attr = ("test", None)
        base.attr_check(attr, self.ds, priority, rv1)
        assert rv1[0] == base.Result(priority, False, "test", ["test not present"])
        # test with empty string
        self.ds.test = ""
        base.attr_check(attr, self.ds, priority, rv2)
        assert rv2[0] == base.Result(
            priority,
            False,
            "test",
            ["test is empty or completely whitespace"],
        )
        # test with whitespace in the form of a space and a tab
        self.ds.test = " 	"
        base.attr_check(attr, self.ds, priority, rv3)
        assert rv3[0] == base.Result(
            priority,
            False,
            "test",
            ["test is empty or completely whitespace"],
        )
        # test with actual string contents
        self.ds.test = "abc 123"
        base.attr_check(attr, self.ds, priority, rv4)
        assert rv4[0] == base.Result(priority, True, "test", [])

    def test_attr_in_valid_choices(self):
        """Tests attribute membership in a set"""
        rv1, rv2, rv3 = [], [], []
        priority = base.BaseCheck.MEDIUM
        valid_choices = ["a", "b", "c"]
        attr = ("test", valid_choices)
        base.attr_check(attr, self.ds, priority, rv1)
        assert rv1[0] == base.Result(priority, (0, 2), "test", ["test not present"])
        self.ds.test = ""
        base.attr_check(attr, self.ds, priority, rv2)
        assert rv2[0] == base.Result(
            priority,
            (1, 2),
            "test",
            [f"test present, but not in expected value list ({valid_choices})"],
        )
        self.ds.test = "a"
        base.attr_check(attr, self.ds, priority, rv3)
        assert rv3[0] == base.Result(priority, (2, 2), "test", [])

    def test_attr_fn(self):
        """Test attribute against a checker function"""

        # simple test.  In an actual program, this use case would be covered
        rv1, rv2, rv3 = [], [], []
        priority = base.BaseCheck.MEDIUM

        def verify_dummy(ds):
            """Sample function that will be called when passed into attr_check"""
            try:
                if ds.dummy + "y" == "dummyy":
                    return base.ratable_result(True, "dummy", [])
                else:
                    return base.ratable_result(False, "dummy", [ds.dummy + "y"])
            except AttributeError:
                return base.ratable_result(False, "dummy", [])

        attr = ("dummy", verify_dummy)
        base.attr_check(attr, self.ds, priority, rv1)
        assert rv1[0] == base.Result(priority, False, "dummy", [])
        self.ds.dummy = "doomy"
        base.attr_check(attr, self.ds, priority, rv2)
        assert rv2[0] == base.Result(priority, False, "dummy", ["doomyy"])
        self.ds.dummy = "dummy"
        base.attr_check(attr, self.ds, priority, rv3)
        assert rv3[0] == base.Result(priority, True, "dummy", [])

    def test_get_test_ctx(self):
        # acdd refers to a BaseCheck instance here -- perhaps the variable name
        # should reflect that?
        ctx = self.acdd.get_test_ctx(base.BaseCheck.HIGH, "Dummy Name")
        ctx.assert_true(1 + 1 == 2, "One plus one equals two")
        assert ctx.out_of == 1
        assert ctx.messages == []

        # ctx2 should be receive the same test context
        ctx2 = self.acdd.get_test_ctx(base.BaseCheck.HIGH, "Dummy Name")
        assert ctx is ctx2
        # will fail, obviously
        ctx2.assert_true(1 + 1 == 3, "One plus one equals three")
        assert ctx.out_of == 2
        assert ctx2.out_of == 2
        assert ctx2.messages == ["One plus one equals three"]

        ctx2 = self.acdd.get_test_ctx(base.BaseCheck.HIGH, "Test Name", "test_var_name")
        ctx3 = self.acdd.get_test_ctx(base.BaseCheck.HIGH, "Test Name", "test_var_name")
        # check that variable cache is working
        assert ctx3 is (
            self.acdd._defined_results["Test Name"]["test_var_name"][
                base.BaseCheck.HIGH
            ]
        )

    def test_email_validation(self):
        test_attr_name = "test"
        validator = base.EmailValidator()
        assert validator.validate(test_attr_name, "foo@bar.com")[0]
        bad_result = validator.validate(test_attr_name, "foo@@bar.com")
        assert not bad_result[0]
        assert bad_result[1] == ["test must be a valid email address"]

    def test_url_validation(self):
        """
        Test that URL validation works properly
        """
        test_attr_name = "test"
        # invalid URL
        test_url = "ssh://invalid_url"
        validator = base.UrlValidator()
        bad_result = validator.validate(test_attr_name, test_url)
        assert not bad_result[0]
        assert bad_result[1] == ["test must be a valid URL"]
        # valid URL
        test_url = "https://ioos.us"
        assert validator.validate(test_attr_name, test_url)[0]
        # test with CSV splitting rules, including checks with embedded commas,
        # which can appear in parts of URLs
        validator = base.UrlValidator(base.csv_splitter)
        url_multi_string = '"http://some-scientific-site.com/depth,temp",https://ioos.us/,http://google.com'
        assert validator.validate(test_attr_name, url_multi_string)[0]
        # add something that's invalid as a URL and check
        url_multi_string += ",noaa.ioos.webmaster@noaa.gov"
        bad_result = validator.validate(test_attr_name, url_multi_string)
        assert not bad_result[0]
        assert bad_result[1] == ["test must be a valid URL"]


class TestGenericFile:
    """
    Tests the GenericFile class.
    """

    def test_create_generic_file_success(self):
        path = "/tmp/test.txt"
        gf = base.GenericFile(path)
        assert gf.filepath() == path

    def test_create_generic_file_failure(self):
        gf = base.GenericFile("will not match")
        assert gf.filepath() != "do NOT MATCH"



import os

import numpy as np
from netCDF4 import Dataset

from compliance_checker.acdd import ACDD1_1Check, ACDD1_3Check
from compliance_checker.tests import BaseTestCase
from compliance_checker.tests.helpers import MockTimeSeries
from compliance_checker.tests.resources import STATIC_FILES


def to_singleton_var(item):
    """
    Get the first value of a list if this implements iterator protocol and is
    not a string
    """
    return [
        x[0] if hasattr(x, "__iter__") and not isinstance(x, str) else x for x in item
    ]


def check_varset_nonintersect(group0, group1):
    """
    Returns true if both groups contain the same elements, regardless of
    order.

    :param list group0: A list of strings to compare
    :param list group1: A list of strings to compare
    """
    # Performs symmetric difference on two lists converted to sets
    return len(set(group0) ^ set(group1)) == 0


class TestACDD1_1(BaseTestCase):
    # TODO superclass this so ACDD1_3 can inherit?
    # Adapted using `pandas.read_html` from URL
    # http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery_1-1
    expected = {
        "Highly Recommended": ["title", "summary", "keywords"],
        "Highly Recommended Variable Attributes": [
            "long_name",
            "standard_name",
            "units",
            "coverage_content_type",
        ],
        "Recommended": [
            "id",
            "naming_authority",
            "keywords_vocabulary",
            "history",
            "comment",
            "date_created",
            "creator_name",
            "creator_url",
            "creator_email",
            "institution",
            "project",
            "processing_level",
            "geospatial_bounds",
            "geospatial_lat_min",
            "geospatial_lat_max",
            "geospatial_lon_min",
            "geospatial_lon_max",
            "geospatial_vertical_min",
            "geospatial_vertical_max",
            "time_coverage_start",
            "time_coverage_end",
            "time_coverage_duration",
            "time_coverage_resolution",
            "standard_name_vocabulary",
            "license",
        ],
        "Suggested": [
            "contributor_name",
            "contributor_role",
            "publisher_name",
            "publisher_url",
            "publisher_email",
            "date_modified",
            "date_issued",
            "geospatial_lat_units",
            "geospatial_lat_resolution",
            "geospatial_lon_units",
            "geospatial_lon_resolution",
            "geospatial_vertical_units",
            "geospatial_vertical_resolution",
            "geospatial_vertical_positive",
        ],
    }

    def setup_method(self):
        # Use the NCEI Gold Standard Point dataset for ACDD checks
        self.ds = self.load_dataset(STATIC_FILES["ncei_gold_point_1"])
        self.acdd = ACDD1_1Check()
        self.acdd_highly_recommended = to_singleton_var(self.acdd.high_rec_atts)
        self.acdd_recommended = to_singleton_var(self.acdd.rec_atts)
        self.acdd_suggested = to_singleton_var(self.acdd.sug_atts)

    def test_cc_meta(self):
        assert self.acdd._cc_spec == "acdd"
        assert self.acdd._cc_spec_version == "1.1"

    def test_highly_recommended(self):
        """
        Checks that all highly recommended attributes are present
        """
        assert check_varset_nonintersect(
            self.expected["Highly Recommended"],
            self.acdd_highly_recommended,
        )

        # Check the reference dataset, NCEI 1.1 Gold Standard Point
        missing = ["\"Conventions\" does not contain 'ACDD-1.3'"]
        results = self.acdd.check_high(self.ds)
        for result in results:
            if result.msgs and all(m in missing for m in result.msgs):
                # only the Conventions check should have failed
                self.assert_result_is_bad(result)
            self.assert_result_is_good(result)

        # Create an empty dataset that writes to /dev/null This acts as a
        # temporary netCDF file in-memory that never gets written to disk.
        # TODO: DRY these dataset open/close, possibly with pytest features
        empty_ds = Dataset(os.devnull, "w", memory=True, diskless=True)
        results = self.acdd.check_high(empty_ds)
        for result in results:
            self.assert_result_is_bad(result)
        empty_ds.close()

    def test_recommended(self):
        """
        Checks that all recommended attributes are present
        """
        # 'geospatial_bounds' attribute currently has its own separate check
        # from the list of required atts
        assert check_varset_nonintersect(
            self.expected["Recommended"],
            self.acdd_recommended,
        )

        ncei_exceptions = [
            "geospatial_bounds not present",
            "time_coverage_duration not present",
            "time_coverage_resolution not present",
        ]
        results = self.acdd.check_recommended(self.ds)
        for result in results:
            if (result.msgs) and all(
                m in ncei_exceptions for m in result.msgs
            ):  # we're doing string comparisons, this is kind of hacky...
                self.assert_result_is_bad(result)
                continue

            self.assert_result_is_good(result)

        # Create an empty dataset that writes to /dev/null This acts as a
        # temporary netCDF file in-memory that never gets written to disk.
        empty_ds = Dataset(os.devnull, "w", diskless=True, memory=True)

        results = self.acdd.check_recommended(empty_ds)
        for result in results:
            self.assert_result_is_bad(result)
        empty_ds.close()

    def test_suggested(self):
        """
        Checks that all suggested attributes are present
        """
        assert check_varset_nonintersect(
            self.expected["Suggested"],
            self.acdd_suggested,
        )

        # Attributes that are missing from NCEI but should be there
        missing = [
            "geospatial_lat_resolution not present",
            "geospatial_lon_resolution not present",
            "geospatial_vertical_resolution not present",
        ]

        results = self.acdd.check_suggested(self.ds)
        for result in results:
            if (result.msgs) and all(
                m in missing for m in result.msgs
            ):  # we're doing string comparisons, this is kind of hacky...
                self.assert_result_is_bad(result)
                continue

            self.assert_result_is_good(result)

        # Create an empty dataset that writes to /dev/null This acts as a
        # temporary netCDF file in-memory that never gets written to disk.
        empty_ds = Dataset(os.devnull, "w", diskless=True, memory=True)

        results = self.acdd.check_recommended(empty_ds)
        for result in results:
            self.assert_result_is_bad(result)
        empty_ds.close()

    def test_acknowldegement_check(self):
        """Check both the British and American English spellings of 'acknowledgement'"""
        # Check British Spelling
        try:
            empty0 = Dataset(os.devnull, "w", diskless=True, memory=True)
            result = self.acdd.check_acknowledgment(empty0)
            self.assert_result_is_bad(result)

            empty0.acknowledgement = "Attribution goes here"
            result = self.acdd.check_acknowledgment(empty0)
            self.assert_result_is_good(result)
        finally:
            empty0.close()

        try:
            # Check American spelling
            empty1 = Dataset(os.devnull, "w", diskless=True, memory=True)
            result = self.acdd.check_acknowledgment(empty1)
            self.assert_result_is_bad(result)

            empty1.acknowledgment = "Attribution goes here"
            result = self.acdd.check_acknowledgment(empty1)
            self.assert_result_is_good(result)
        finally:
            empty1.close()


class TestACDD1_3(BaseTestCase):
    # Adapted using `pandas.read_html` from URL
    # http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery_1-3
    expected = {
        "Suggested": [
            "creator_type",
            "creator_institution",
            "publisher_type",
            "publisher_institution",
            "program",
            "contributor_name",
            "contributor_role",
            "geospatial_lat_units",
            "geospatial_lat_resolution",
            "geospatial_lon_units",
            "geospatial_lon_resolution",
            "geospatial_vertical_units",
            "geospatial_vertical_resolution",
            "date_modified",
            "date_issued",
            "date_metadata_modified",
            "product_version",
            "keywords_vocabulary",
            "platform",
            "platform_vocabulary",
            "instrument",
            "instrument_vocabulary",
            "metadata_link",
            "references",
        ],
        "Highly Recommended": ["title", "summary", "keywords", "Conventions"],
        "Recommended": [
            "id",
            "naming_authority",
            "history",
            "source",
            "processing_level",
            "comment",
            "license",
            "standard_name_vocabulary",
            "date_created",
            "creator_name",
            "creator_email",
            "creator_url",
            "institution",
            "project",
            "publisher_name",
            "publisher_email",
            "publisher_url",
            "geospatial_bounds",
            "geospatial_bounds_crs",
            "geospatial_bounds_vertical_crs",
            "geospatial_lat_min",
            "geospatial_lat_max",
            "geospatial_lon_min",
            "geospatial_lon_max",
            "geospatial_vertical_min",
            "geospatial_vertical_max",
            "geospatial_vertical_positive",
            "time_coverage_start",
            "time_coverage_end",
            "time_coverage_duration",
            "time_coverage_resolution",
        ],
        "Highly Recommended Variable Attributes": [
            "long_name",
            "standard_name",
            "units",
            "coverage_content_type",
        ],
    }

    def setup_method(self):
        # Use the NCEI Gold Standard Point dataset for ACDD checks
        self.ds = self.load_dataset(STATIC_FILES["ncei_gold_point_2"])
        self.acdd = ACDD1_3Check()
        self.acdd_highly_recommended = to_singleton_var(self.acdd.high_rec_atts)
        self.acdd_recommended = to_singleton_var(self.acdd.rec_atts)
        self.acdd_suggested = to_singleton_var(self.acdd.sug_atts)

    def test_cc_meta(self):
        assert self.acdd._cc_spec == "acdd"
        assert self.acdd._cc_spec_version == "1.3"

    def test_highly_recommended(self):
        """
        Checks that all highly recommended attributes are present
        """
        assert check_varset_nonintersect(
            self.expected["Highly Recommended"],
            self.acdd_highly_recommended,
        )

        results = self.acdd.check_high(self.ds)
        for result in results:
            # NODC 2.0 has a different value in the conventions field
            self.assert_result_is_good(result)

    def test_recommended(self):
        """
        Checks that all recommended attributes are present
        """
        assert check_varset_nonintersect(
            self.expected["Recommended"],
            self.acdd_recommended,
        )

        results = self.acdd.check_recommended(self.ds)
        ncei_exceptions = [
            "time_coverage_duration not present",
            "time_coverage_resolution not present",
        ]
        for result in results:
            if (result.msgs) and all(
                m in ncei_exceptions for m in result.msgs
            ):  # we're doing string comparisons, this is kind of hacky...
                self.assert_result_is_bad(result)
                continue

            self.assert_result_is_good(result)

    def test_suggested(self):
        """
        Checks that all suggested attributes are present
        """
        assert check_varset_nonintersect(
            self.expected["Suggested"],
            self.acdd_suggested,
        )

        results = self.acdd.check_suggested(self.ds)
        # NCEI does not require or suggest resolution attributes
        ncei_exceptions = [
            "geospatial_lat_resolution not present",
            "geospatial_lon_resolution not present",
            "geospatial_vertical_resolution not present",
        ]
        for result in results:
            if (result.msgs) and all(
                m in ncei_exceptions for m in result.msgs
            ):  # we're doing string comparisons, this is kind of hacky...
                self.assert_result_is_bad(result)
                continue
            self.assert_result_is_good(result)

    def test_variables(self):
        """
        Test that variables are checked for required attributes
        """
        # Create an empty dataset that writes to /dev/null This acts as a
        # temporary netCDF file in-memory that never gets written to disk.
        empty_ds = Dataset(os.devnull, "w", diskless=True, memory=True)
        # self.addCleanup(empty_ds.close)

        # The dataset needs at least one variable to check that it's missing
        # all the required attributes.
        empty_ds.createDimension("time", 1)
        empty_ds.createVariable("fake", "float32", ("time",))

        # long_name
        results = self.acdd.check_var_long_name(self.ds)
        for result in results:
            self.assert_result_is_good(result)

        results = self.acdd.check_var_long_name(empty_ds)
        assert len(results) == 1
        for result in results:
            self.assert_result_is_bad(result)

        # standard_name
        results = self.acdd.check_var_standard_name(self.ds)
        for result in results:
            self.assert_result_is_good(result)

        results = self.acdd.check_var_standard_name(empty_ds)
        assert len(results) == 1
        for result in results:
            self.assert_result_is_bad(result)

        # units
        results = self.acdd.check_var_units(self.ds)
        for result in results:
            self.assert_result_is_good(result)

        results = self.acdd.check_var_units(empty_ds)
        assert len(results) == 1
        for result in results:
            self.assert_result_is_bad(result)

        empty_ds.close()

    def test_vertical_extents(self):
        """
        Test vertical extents are being checked
        """

        result = self.acdd.check_vertical_extents(self.ds)
        self.assert_result_is_good(result)

    def test_geospatial_bounds(self):
        """
        Test geospatial bounds are checked and provide a good error message
        """
        # Create an empty dataset that writes to /dev/null This acts as a
        # temporary netCDF file in-memory that never gets written to disk.
        empty_ds = Dataset(os.devnull, "w", diskless=True)

        # Misspelled WKT. Error message should include the attribute checked
        # and the value that was provided for easy troubleshooting
        empty_ds.geospatial_bounds = "POIT (-123.458000 38.048000)"
        results = self.acdd.check_recommended(empty_ds)
        for result in results:
            if result.variable_name == "geospatial_bounds":
                assert (
                    "Could not parse WKT from geospatial_bounds,"
                    f' possible bad value: "{empty_ds.geospatial_bounds}"'
                    in result.msgs
                )
        empty_ds.close()

    def test_time_extents(self):
        """
        Test that the time extents are being checked
        """
        result = self.acdd.check_time_extents(self.ds)
        self.assert_result_is_good(result)

        empty_ds = Dataset(os.devnull, "w", diskless=True)

        # The dataset needs at least one variable to check that it's missing
        # all the required attributes.
        empty_ds.createDimension("time", 1)
        time_var = empty_ds.createVariable("time", "float32", ("time",))
        time_var.units = "seconds since 1970-01-01 00:00:00 UTC"
        time_var[:] = [1451692800]  # 20160102T000000Z in seconds since epoch
        empty_ds.time_coverage_start = "20160102T000000Z"
        empty_ds.time_coverage_end = "20160102T000000Z"
        result = self.acdd.check_time_extents(empty_ds)
        self.assert_result_is_good(result)
        # try the same thing with time offsets
        time_var.units = "seconds since 1970-01-01 00:00:00-10:00"
        empty_ds.time_coverage_start = "20160102T000000-1000"
        empty_ds.time_coverage_end = "20160102T000000-1000"
        result = self.acdd.check_time_extents(empty_ds)
        self.assert_result_is_good(result)
        empty_ds.close()

    def test_check_lat_extents(self):
        """Test the check_lat_extents() method behaves expectedly"""

        # create dataset using MockDataset, give it lat/lon dimensions
        ds = MockTimeSeries()
        ds.variables["lat"][:] = np.linspace(
            -135.0,
            -130.0,
            num=500,
        )  # arbitrary, but matches time dim size

        # test no values, expect failure
        result = self.acdd.check_lat_extents(ds)
        self.assert_result_is_bad(result)

        # give integer geospatial_lat_max/min, test
        ds.setncattr("geospatial_lat_min", -135)
        ds.setncattr("geospatial_lat_max", -130)

        result = self.acdd.check_lat_extents(ds)
        self.assert_result_is_good(result)

        # give float geospatial_lat_min/max, test
        ds.setncattr("geospatial_lat_min", -135.0)
        ds.setncattr("geospatial_lat_max", -130.0)

        result = self.acdd.check_lat_extents(ds)
        self.assert_result_is_good(result)

        # give string (in number-form), test
        ds.setncattr("geospatial_lat_min", "-135.")
        ds.setncattr("geospatial_lat_max", "-130.")

        result = self.acdd.check_lat_extents(ds)
        self.assert_result_is_good(result)

        # give garbage string -- expect failure
        ds.setncattr("geospatial_lat_min", "bad")
        ds.setncattr("geospatial_lat_max", "val")

        result = self.acdd.check_lat_extents(ds)
        self.assert_result_is_bad(result)

    def test_check_lon_extents(self):
        """Test the check_lon_extents() method behaves expectedly"""

        # create dataset using MockDataset, give it lat/lon dimensions
        ds = MockTimeSeries()
        ds.variables["lon"][:] = np.linspace(65.0, 67.0, num=500)

        # test no values, expect failure
        # result = self.acdd.check_lon_extents(ds)
        # self.assert_result_is_bad(result)

        # give integer geospatial_lon_max/min, test
        ds.setncattr("geospatial_lon_min", 65)
        ds.setncattr("geospatial_lon_max", 67)

        result = self.acdd.check_lon_extents(ds)
        self.assert_result_is_good(result)

        # give float geospatial_lon_min/max, test
        ds.setncattr("geospatial_lon_min", 65.0)
        ds.setncattr("geospatial_lon_max", 67.0)

        result = self.acdd.check_lon_extents(ds)
        self.assert_result_is_good(result)

        # give string (in number-form), test
        ds.setncattr("geospatial_lon_min", "65.")
        ds.setncattr("geospatial_lon_max", "67.")

        result = self.acdd.check_lon_extents(ds)
        self.assert_result_is_good(result)

        # give garbage string -- expect failure
        ds.setncattr("geospatial_lon_min", "bad")
        ds.setncattr("geospatial_lon_max", "val")

        result = self.acdd.check_lon_extents(ds)
        self.assert_result_is_bad(result)

    def test_check_geospatial_vertical_max(self):
        ds = MockTimeSeries()
        ds.variables["depth"][:] = np.linspace(0.0, 30.0, num=500)

        # give integer geospatial_vertical_max/min, test
        ds.setncattr("geospatial_vertical_min", 0)
        ds.setncattr("geospatial_vertical_max", 30)

        result = self.acdd.check_vertical_extents(ds)
        self.assert_result_is_good(result)

        # give float geospatial_vertical_min/max, test
        ds.setncattr("geospatial_vertical_min", 0.0)
        ds.setncattr("geospatial_vertical_max", 30.0)

        result = self.acdd.check_vertical_extents(ds)
        self.assert_result_is_good(result)

        # give string (in number-form), test
        ds.setncattr("geospatial_vertical_min", "0.")
        ds.setncattr("geospatial_vertical_max", "30.")

        result = self.acdd.check_vertical_extents(ds)
        self.assert_result_is_good(result)

        # give garbage string -- expect failure
        ds.setncattr("geospatial_vertical_min", "bad")
        ds.setncattr("geospatial_vertical_max", "val")

        result = self.acdd.check_vertical_extents(ds)
        self.assert_result_is_bad(result)

        # all masked values mean that there are no valid array elements to get
        # the min/max of
        ds.setncattr("geospatial_vertical_min", 0.0)
        ds.setncattr("geospatial_vertical_max", 30.0)
        ds.variables["depth"][:] = np.ma.masked_all(ds.variables["depth"].shape)
        result = self.acdd.check_vertical_extents(ds)
        self.assert_result_is_bad(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import compliance_checker.cf.util as cfutil  # noqa: F401
from compliance_checker.base import (  # noqa: F401
    BaseCheck,
    BaseNCCheck,
    Result,
    TestCtx,
)
from compliance_checker.cf import util  # noqa: F401
from compliance_checker.cf.appendix_d import (  # noqa: F401
    dimless_vertical_coordinates_1_6,
    dimless_vertical_coordinates_1_7,
    no_missing_terms,
)
from compliance_checker.cf.appendix_e import cell_methods16  # noqa: F401
from compliance_checker.cf.appendix_e import cell_methods17  # noqa: F401
from compliance_checker.cf.appendix_f import ellipsoid_names17  # noqa: F401
from compliance_checker.cf.appendix_f import (  # noqa: F401
    grid_mapping_attr_types16,
    grid_mapping_attr_types17,
    grid_mapping_dict16,
    grid_mapping_dict17,
    horizontal_datum_names17,
    prime_meridian_names17,
)

# Version specific checkers organized in other modules
from compliance_checker.cf.cf_1_6 import CF1_6Check  # noqa: F401
from compliance_checker.cf.cf_1_7 import CF1_7Check  # noqa: F401
from compliance_checker.cf.cf_1_8 import CF1_8Check  # noqa: F401
from compliance_checker.cf.cf_1_9 import CF1_9Check  # noqa: F401
from compliance_checker.cf.cf_1_10 import CF1_10Check  # noqa: F401
from compliance_checker.cf.cf_1_11 import CF1_11Check  # noqa: F401



#!/usr/bin/env python
"""
compliance_checker/protocols/opendap.py

Functions to assist in determining if the URL is an OPeNDAP endpoint
"""
import urllib.parse
import urllib.request

import requests


def create_DAP_variable_str(url):
    """
    Create a URL-encoded string of variables for a given DAP dataset.
    Works on OPeNDAP datasets.

    Parameters
    ----------
    url (str): endpoint to *DAP dataset

    Returns
    -------
    str
    """

    # get dds
    with urllib.request.urlopen(f"{url}.dds") as resp:
        _str = resp.read().decode()[8:]

    # remove beginning and ending braces, split on newlines
    no_braces_newlines = list(
        filter(lambda x: "{" not in x and "}" not in x, _str.split("\n")),
    )

    # remove all the extra space used in the DDS string
    no_spaces = list(filter(None, (x.strip(" ") for x in no_braces_newlines)))

    # now need to split from type, grab only the variable and remove ;
    vars_only = [x.split(" ")[-1].strip(";") for x in no_spaces]

    # encode as proper URL characters
    varstr = urllib.parse.quote(",".join(vars_only))
    return varstr


def is_opendap(url):
    """
    Returns True if the URL is a valid OPeNDAP URL

    :param str url: URL for a remote OPeNDAP endpoint
    """
    # If the server replies to a Data Attribute Structure request
    if url.endswith("#fillmismatch"):
        das_url = url.replace("#fillmismatch", ".das")
    else:
        das_url = url + ".das"

    try:
        response = requests.get(das_url, allow_redirects=True)

        if "xdods-server" in response.headers:
            return True
        # Check if it is an access restricted ESGF thredds service
        if (
            response.status_code == 401
            and "text/html" in response.headers["content-type"]
            and "The following URL requires authentication:" in response.text
        ):
            return True
    except requests.exceptions.InvalidSchema:
        return False  # not opendap if url + ".das" isn't found
    return False



#!/usr/bin/env python

import pytest

from compliance_checker.cf import util

# get current std names table version (it changes)
std_names = util.StandardNameTable()

# The meat and potatoes of most integration tests lies here
# To add a new integration test, simply add a new .CDL dataset somewhere in tests/data,
# and append this list with another tuple in the form:
# (file stem to test dataset in tests/data or subfolder : str, [expected messages]: list of strs )
# Note that this will also assert there was a missed compliance check with your dataset:
# assert scored < out_of
dataset_stem__expected_messages = [
    (
        "sldmb_43093_agg",
        [
            "attribute time:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            "attribute lat:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            "attribute lon:_CoordianteAxisType should begin with a letter and be composed of letters, digits, and underscores",
            "§2.6.2 global attribute history should exist and be a non-empty string",
            f"standard_name temperature is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['air_temperature', 'soil_temperature', 'snow_temperature']",
            "temperature's auxiliary coordinate specified by the coordinates attribute, precise_lat, is not a variable in this dataset",
            "temperature's auxiliary coordinate specified by the coordinates attribute, precise_lon, is not a variable in this dataset",
        ],
    ),
    (
        "usgs_dem_saipan",
        ['§2.6.1 Conventions global attribute does not contain "CF-1.9"'],
    ),
    (
        "l01-met",
        [
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'air_temperature_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'barometric_pressure_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.float64'>) and parent variable 'wind_gust' (type: <class 'numpy.float32'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'wind_gust_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.float64'>) and parent variable 'wind_speed' (type: <class 'numpy.float32'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'wind_speed_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.float64'>) and parent variable 'wind_direction' (type: <class 'numpy.float32'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'wind_direction_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            "Attribute 'valid_range' (type: <class 'numpy.int16'>) and parent variable 'visibility_qc' (type: <class 'numpy.int8'>) must have equivalent datatypes",
            '§2.6.1 Conventions global attribute does not contain "CF-1.8"',
            f"standard_name visibility is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['visibility_in_air']",
            'Standard name modifier "data_quality" for variable visibility_qc is not a valid modifier according to CF Appendix C',
            f"standard_name wind_direction is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['wind_to_direction', 'wind_from_direction', 'wind_gust_from_direction']",
            'Standard name modifier "data_quality" for variable wind_direction_qc is not a valid modifier according to CF Appendix C',
            f"standard_name wind_gust is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['y_wind_gust', 'x_wind_gust', 'wind_speed_of_gust']",
            'Standard name modifier "data_quality" for variable wind_gust_qc is not a valid modifier according to CF Appendix C',
            'Standard name modifier "data_quality" for variable air_temperature_qc is not a valid modifier according to CF Appendix C',
            f"standard_name use_wind is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['y_wind', 'x_wind']",
            f"standard_name barometric_pressure is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['air_pressure', 'reference_pressure', 'barometric_altitude']",
            'Standard name modifier "data_quality" for variable barometric_pressure_qc is not a valid modifier according to CF Appendix C',
            'Standard name modifier "data_quality" for variable wind_speed_qc is not a valid modifier according to CF Appendix C',
            f"standard_name barometric_pressure is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['air_pressure', 'reference_pressure', 'barometric_altitude']",
            "CF recommends latitude variable 'lat' to use units degrees_north",
            "CF recommends longitude variable 'lon' to use units degrees_east",
        ],
    ),
    ("sp041", ["lat_qc is not a variable in this dataset"]),
    (
        "3mf07",
        [
            "latitude:valid_min must be a numeric type not a string",
            "latitude:valid_max must be a numeric type not a string",
            "longitude:valid_min must be a numeric type not a string",
            "longitude:valid_max must be a numeric type not a string",
            "§2.6.2 references global attribute should be a non-empty string",
            "§2.6.2 comment global attribute should be a non-empty string",
            "dimensions for auxiliary coordinate variable z (z) are not a subset of dimensions for variable flag (profile)",
            "dimensions for auxiliary coordinate variable z (z) are not a subset of dimensions for variable haul (profile)",
        ],
    ),
    (
        "ooi_glider",
        [
            "§2.6.2 comment global attribute should be a non-empty string",
            "Attribute long_name or/and standard_name is highly recommended for variable deployment",
            "latitude variable 'latitude' should define standard_name='latitude' or axis='Y'",
            "longitude variable 'longitude' should define standard_name='longitude' or axis='X'",
        ],
    ),
    (
        "swan",
        [
            "global attribute _CoordSysBuilder should begin with a letter and be composed of letters, digits, and underscores",
            '§2.6.1 Conventions global attribute does not contain "CF-1.9"',
            'Units "hours since 2013-02-18T00:00:00Z" for variable time_offset must be convertible to canonical units "s"',
            "lon's axis attribute must be T, X, Y, or Z, currently x",
            "lat's axis attribute must be T, X, Y, or Z, currently y",
            "z's axis attribute must be T, X, Y, or Z, currently z",
            "z: vertical coordinates not defining pressure must include a positive attribute that is either 'up' or 'down'",
            "GRID is not a valid CF featureType. It must be one of point, timeseries, trajectory, profile, timeseriesprofile, trajectoryprofile",
        ],
    ),
    (
        "kibesillah",
        ["§2.6.2 global attribute title should exist and be a non-empty string"],
    ),
    (
        "pr_inundation",
        [
            'Units "degrees2" for variable area must be convertible to canonical units "m2"',
            "waterlevel's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "velocity_x's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), Layer (Z), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "velocity_y's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), Layer (Z), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "tau_x's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "tau_y's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are time (T), m (A), n (A) (with U: other/unknown; L: unlimited).",
            "§2.6.2 grid_depth:comment should be a non-empty string",
            "§2.6.2 depth:comment should be a non-empty string",
            "§2.6.2 institution global attribute should be a non-empty string",
            "§2.6.2 comment global attribute should be a non-empty string",
            'Units "degrees2" for variable area must be convertible to canonical units "m2"',
            "k: vertical coordinates not defining pressure must include a positive attribute that is either 'up' or 'down'",
            "grid_longitude has no coordinate associated with a variable identified as true latitude/longitude; its coordinate variable should also share a subset of grid_longitude's dimensions",
            "grid_latitude has no coordinate associated with a variable identified as true latitude/longitude; its coordinate variable should also share a subset of grid_latitude's dimensions",
            "time_bounds might be a cell boundary variable but there are no variables that define it as a boundary using the `bounds` attribute.",
        ],
    ),
    (
        "ww3",
        [
            "§2.6.2 global attribute title should exist and be a non-empty string",
            "§2.6.2 global attribute history should exist and be a non-empty string",
            "§2.6.1 Conventions field is not present",
            "Attribute long_name or/and standard_name is highly recommended for variable time",
            "Attribute long_name or/and standard_name is highly recommended for variable lon",
            "Attribute long_name or/and standard_name is highly recommended for variable lat",
            "latitude variable 'lat' should define standard_name='latitude' or axis='Y'",
            "longitude variable 'lon' should define standard_name='longitude' or axis='X'",
        ],
    ),
    (
        "glcfs",
        [
            # TODO: referenced/relative time is treated like time units
            'Units "hours since 2016-01-01T12:00:00Z" for variable time_offset must be convertible to canonical units "s"',
            f"standard_name cloud_cover is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['land_cover', 'land_cover_lccs', 'cloud_albedo']",
            f"standard_name dew_point is not defined in Standard Name Table v{std_names._version}. Possible close match(es): ['dew_point_depression', 'dew_point_temperature']",
            (
                "GRID is not a valid CF featureType. It must be one of point, timeseries, "
                "trajectory, profile, timeseriesprofile, trajectoryprofile"
            ),
            (
                "global attribute _CoordSysBuilder should begin with a letter and "
                "be composed of letters, digits, and underscores"
            ),
            'units for cl, "fraction" are not recognized by UDUNITS',
        ],
    ),
    (
        "bad_cf_role",
        [
            "§2.6.2 global attribute title should exist and be a non-empty string",
            "§2.6.2 global attribute history should exist and be a non-empty string",
            "§2.6.1 Conventions field is not present",
            "§9.5 The only acceptable values of cf_role for Discrete Geometry CF data sets are timeseries_id, profile_id, and trajectory_id",
        ],
    ),
    pytest.param(
        "ocos",
        [
            "AKs's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "AKt's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "AKv's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "latent's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "lwrad's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "salt's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "sensible's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "shflux's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "swrad's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "temp's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "tke's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "u's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_u (A), xi_u (A) (with U: other/unknown; L: unlimited).",
            "ubar's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_u (A), xi_u (A) (with U: other/unknown; L: unlimited).",
            "v's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_rho (Z), eta_v (A), xi_v (A) (with U: other/unknown; L: unlimited).",
            "vbar's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_v (A), xi_v (A) (with U: other/unknown; L: unlimited).",
            "w's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), s_w (Z), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            "zeta's spatio-temporal dimensions are not in the recommended order T, Z, Y, X and/or further dimensions are not located left of T, Z, Y, X. The dimensions (and their guessed types) are ocean_time (T), eta_rho (A), xi_rho (A) (with U: other/unknown; L: unlimited).",
            '§2.6.1 Conventions global attribute does not contain "CF-1.8"',
            "CF recommends latitude variable 'lat_rho' to use units degrees_north",
            "CF recommends latitude variable 'lat_u' to use units degrees_north",
            "CF recommends latitude variable 'lat_v' to use units degrees_north",
            "CF recommends latitude variable 'lat_psi' to use units degrees_north",
            "CF recommends longitude variable 'lon_rho' to use units degrees_east",
            "CF recommends longitude variable 'lon_u' to use units degrees_east",
            "CF recommends longitude variable 'lon_v' to use units degrees_east",
            "CF recommends longitude variable 'lon_psi' to use units degrees_east",
            "§4.3.3 The standard_name of `s_rho` must map to the correct computed_standard_name, `['altitude', 'height_above_geopotential_datum', 'height_above_mean_sea_level', 'height_above_reference_ellipsoid']`",
            "§4.3.3 The standard_name of `s_w` must map to the correct computed_standard_name, `['altitude', 'height_above_geopotential_datum', 'height_above_mean_sea_level', 'height_above_reference_ellipsoid']`",
        ],
        marks=pytest.mark.slowtest,
    ),
]

mult_msgs_diff = "Failed to find the following messages:\n{missing_msgs}\n\n\
        These were the messages captured:\n{found_msgs}\n\
            Please check wording and section names if messages have been altered since this test was written"


class TestCFIntegration:
    # --------------------------------------------------------------------------------
    # Helper Methods
    # --------------------------------------------------------------------------------

    def get_results(self, check_results, checksuite):
        """
        Returns a tuple of the value scored, possible, and a list of messages
        in the result set.\n
        cs: instance of CheckSuite object
        """
        aggregation = checksuite.build_structure(
            "cf",
            check_results["cf"][0],
            "test",
            1,
        )
        out_of = 0
        scored = 0
        results = aggregation["all_priorities"]
        for r in results:
            if isinstance(r.value, tuple):
                out_of += r.value[1]
                scored += r.value[0]
            else:
                out_of += 1
                scored += int(r.value)

        # Store the messages
        messages = []
        for r in results:
            messages.extend(r.msgs)

        return scored, out_of, messages

    # --------------------------------------------------------------------------------

    @pytest.mark.parametrize(
        "loaded_dataset,expected_messages",
        dataset_stem__expected_messages,
        indirect=[
            "loaded_dataset",
        ],  # must be specified to load this param at runtime, instead of at collection
    )
    def test_cf_integration(self, loaded_dataset, expected_messages, cs):
        # check_results = cs.run(loaded_dataset, [], "cf")
        check_results = cs.run_all(loaded_dataset, ["cf"], skip_checks=[])
        scored, out_of, messages = self.get_results(check_results, cs)

        assert scored < out_of

        assert all(m in messages for m in expected_messages), mult_msgs_diff.format(
            missing_msgs="\n".join([m for m in expected_messages if m not in messages]),
            found_msgs="\n".join(messages),
        )

    @pytest.mark.parametrize(
        "loaded_dataset,wrong_message",
        # From Github issue #845:\n
        # CF: incorrect errors for ragged array structure\n
        # \n
        # Structure needs to be correctly identified\n
        # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#_indexed_ragged_array_representation\n
        # https://github.com/ioos/compliance-checker/issues/845
        [
            ("index_ragged2", "are not a subset of dimensions for variable"),
            ("index_ragged2", "Unidentifiable feature for variable"),
        ],
        indirect=["loaded_dataset"],
    )
    def test_no_incorrect_errors(self, cs, loaded_dataset, wrong_message):
        # check_results = cs.run(loaded_dataset, [], True, "cf")
        check_results = cs.run_all(loaded_dataset, ["cf"], skip_checks=[])
        messages = self.get_results(check_results, cs)[-1]

        assert wrong_message not in "".join(messages)

    @pytest.mark.parametrize("loaded_dataset", ["fvcom"], indirect=True)
    def test_fvcom(self, cs, loaded_dataset):
        # check_results = cs.run(loaded_dataset, [], True, "cf")
        check_results = cs.run_all(loaded_dataset, ["cf"], skip_checks=[])
        scored, out_of, messages = self.get_results(check_results, cs)
        assert scored < out_of

        for msg in messages:
            if msg.startswith("dimensions for auxiliary coordinate variable siglay"):
                break
        # it's not clear to me what this is supposed to be doing -- this else clause is outside of the if
        else:
            raise AssertionError(
                '"dimensions for auxiliary coordinate variable siglay (node, siglay) '
                'are not a subset of dimensions for variable u (siglay, nele, time)"'
                " not in messages",
            )
        assert (
            '§2.6.1 Conventions global attribute does not contain "CF-1.8"'
        ) in messages

    @pytest.mark.parametrize(
        "loaded_dataset",
        ["NCEI_profile_template_v2.0_2016-09-22_181835.151325"],
        indirect=True,
    )
    def test_ncei_templates(self, cs, loaded_dataset):
        """
        Tests some of the NCEI NetCDF templates, which usually should get a
        perfect score.
        """
        # check_results = cs.run(loaded_dataset, [], "cf")
        check_results = cs.run_all(loaded_dataset, ["cf"], skip_checks=[])
        scored, out_of, messages = self.get_results(check_results, cs)
        assert scored < out_of



from pathlib import Path

from netCDF4 import Dataset


class BaseTestCase:
    """Base test case for migrating tests away from unittest and towards pytest only"""

    def shortDescription(self):
        return None

    def __repr__(self):
        name = self.id()
        name = name.split(".")
        return "{} ( {} )".format(
            name[-1],
            ".".join(name[:-2]) + ":" + ".".join(name[-2:]),
        )

    __str__ = __repr__

    def assert_result_is_good(self, result):
        if isinstance(result.value, bool):
            assert result.value is True
        else:
            assert result.value[0] == result.value[1]

    def assert_result_is_bad(self, result):
        if isinstance(result.value, bool):
            assert result.value is False
        else:
            assert result.value[0] != result.value[1]

    # @pytest.fixture
    # def _nc_cleanup(self, request)

    def load_dataset(self, nc_dataset):
        """
        Return a loaded NC Dataset for the given path
        """
        if not isinstance(nc_dataset, (str | Path)):
            raise ValueError("nc_dataset should be a valid path")

        nc_dataset = Dataset(nc_dataset, "r", diskless=True)
        # TODO: Fix cleanup
        # self.addCleanup(nc_dataset.close)
        return nc_dataset



import csv
import itertools
import os
import posixpath
import re
import sys
import warnings
from collections import defaultdict
from functools import lru_cache, partial
from importlib.resources import files
from pkgutil import get_data

import requests
from lxml import etree
from netCDF4 import Dataset, Dimension, Group, Variable

from compliance_checker.cfunits import Unit

_UNITLESS_DB = None
_SEA_NAMES = None

# copied from paegan
# paegan may depend on these later
_possiblet = {
    "time",
    "TIME",
    "Time",
    "t",
    "T",
    "ocean_time",
    "OCEAN_TIME",
    "jd",
    "JD",
    "dn",
    "DN",
    "times",
    "TIMES",
    "Times",
    "mt",
    "MT",
    "dt",
    "DT",
}
_possiblez = {
    "depth",
    "DEPTH",
    "depths",
    "DEPTHS",
    "height",
    "HEIGHT",
    "altitude",
    "ALTITUDE",
    "alt",
    "ALT",
    "Alt",
    "Altitude",
    "h",
    "H",
    "s_rho",
    "S_RHO",
    "s_w",
    "S_W",
    "z",
    "Z",
    "siglay",
    "SIGLAY",
    "siglev",
    "SIGLEV",
    "sigma",
    "SIGMA",
    "vertical",
    "VERTICAL",
    "lev",
    "LEV",
    "level",
    "LEVEL",
}
_possiblex = {
    "x",
    "X",
    "lon",
    "LON",
    "xlon",
    "XLON",
    "lonx",
    "lon_u",
    "LON_U",
    "lon_v",
    "LON_V",
    "lonc",
    "LONC",
    "Lon",
    "Longitude",
    "longitude",
    "LONGITUDE",
    "lon_rho",
    "LON_RHO",
    "lon_psi",
    "LON_PSI",
}
_possibley = {
    "y",
    "Y",
    "lat",
    "LAT",
    "ylat",
    "YLAT",
    "laty",
    "lat_u",
    "LAT_U",
    "lat_v",
    "LAT_V",
    "latc",
    "LATC",
    "Lat",
    "Latitude",
    "latitude",
    "LATITUDE",
    "lat_rho",
    "LAT_RHO",
    "lat_psi",
    "LAT_PSI",
}

_possibleaxis = _possiblet | _possiblez | _possiblex | _possibley


_possiblexunits = {
    "degrees_east",
    "degree_east",
    "degrees_E",
    "degree_E",
    "degreesE",
    "degreeE",
}

_possibleyunits = {
    "degrees_north",
    "degree_north",
    "degrees_N",
    "degree_N",
    "degreesN",
    "degreeN",
}

_possibletunits = {
    "day",
    "days",
    "d",
    "hour",
    "hours",
    "hr",
    "hrs",
    "h",
    "year",
    "years",
    "minute",
    "minutes",
    "m",
    "min",
    "mins",
    "second",
    "seconds",
    "s",
    "sec",
    "secs",
}

_possibleaxisunits = _possiblexunits | _possibleyunits | _possibletunits


VALID_LAT_UNITS = {
    "degrees_north",
    "degree_north",
    "degree_n",
    "degrees_n",
    "degreen",
    "degreesn",
}
VALID_LON_UNITS = {
    "degrees_east",
    "degree_east",
    "degree_e",
    "degrees_e",
    "degreee",
    "degreese",
}


# We can't import appendix d without getting circular imports
DIMENSIONLESS_VERTICAL_COORDINATES = {
    "ocean_s_coordinate",
    "ocean_s_coordinate_g1",
    "ocean_s_coordinate_g2",
    "atmosphere_hybrid_sigma_pressure_coordinate",
    "atmosphere_hybrid_height_coordinate",
    "ocean_double_sigma_coordinate",
    "ocean_sigma_z_coordinate",
    "ocean_sigma_coordinate",
    "atmosphere_sigma_coordinate",
    "atmosphere_ln_pressure_coordinate",
    "atmosphere_sleve_coordinate",
}


def attr_membership(attr_val, value_set, attr_type=str, modifier_fn=lambda x: x):
    """
    Helper function passed to netCDF4.Dataset.get_attributes_by_value
    Checks that `attr_val` exists, has the same type as `attr_type`,
    and is contained in `value_set`
    attr_val: The value of the attribute being checked
    attr_type: A type object that the `attr_val` is expected to have the same
               type as.  If the type is not the same, a warning is issued and
               the code attempts to cast `attr_val` to the expected type.
    value_set: The set against which membership for `attr_val` is tested
    modifier_fn: A function to apply to attr_val prior to applying the set
                 membership test

    """
    if attr_val is None:
        return False

    if not isinstance(attr_val, attr_type):
        warnings.warn(
            f"Attribute is of type {type(attr_val)!r}, {attr_type!r} expected. Attempting to cast to expected type.",
            stacklevel=2,
        )
        try:
            # if the expected type is str, try casting to unicode type
            # since str can't be instantiated
            if attr_type is str:
                new_attr_val = str(attr_val)
            else:
                new_attr_val = attr_type(attr_val)
        # catch casting errors
        except (ValueError, UnicodeEncodeError):
            warnings.warn(f"Could not cast to type {attr_type}", stacklevel=2)
            return False
    else:
        new_attr_val = attr_val

    try:
        is_in_set = modifier_fn(new_attr_val) in value_set
    except Exception as e:
        warnings.warn(
            f"Could not apply modifier function {modifier_fn} to value: {e.msg}",
            stacklevel=2,
        )
        return False

    return is_in_set


@lru_cache(128)
def is_dimensionless_standard_name(standard_name_table, standard_name):
    """
    Returns True if the units for the associated standard name are
    dimensionless.  Dimensionless standard names include those that have no
    units and units that are defined as constant units in the CF standard name
    table i.e. '1', or '1e-3'.
    """
    # standard_name must be string, so if it is not, it is *wrong* by default
    if not isinstance(standard_name, str):
        return False
    found_standard_name = standard_name_table.find(
        f".//entry[@id='{standard_name}']",
    )
    if found_standard_name is not None:
        canonical_units = Unit(found_standard_name.find("canonical_units").text)
        return canonical_units.is_dimensionless()
    # if the standard name is not found, assume we need units for the time being
    else:
        return False


def get_sea_names():
    """
    Returns a list of NODC sea names

    source of list: https://www.ncei.noaa.gov/resources/ocean-data-format-codes
    """
    global _SEA_NAMES
    if _SEA_NAMES is None:
        buf = {}
        with open(
            files("compliance_checker") / "data/seanames.csv",
        ) as f:
            reader = csv.reader(f)
            for code, sea_name in reader:
                buf[sea_name] = code
        _SEA_NAMES = buf
    return _SEA_NAMES


def is_unitless(nc, variable):
    """
    Returns true if the variable is unitless

    Note units of '1' are considered whole numbers or parts but still represent
    physical units and not the absence of units.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: Name of the variable
    """
    units = getattr(nc.variables[variable], "units", None)
    return units is None or units == ""


def is_geophysical(nc, variable):
    """
    Returns true if the dataset's variable is likely a geophysical variable

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: Name of the variable
    """
    ncvar = nc.variables[variable]

    if getattr(ncvar, "cf_role", None):
        return False

    # Check for axis
    if getattr(ncvar, "axis", None):
        return False

    standard_name_test = getattr(ncvar, "standard_name", "")
    unitless = is_unitless(nc, variable)

    if not isinstance(standard_name_test, str):
        warnings.warn(
            f"Variable {variable} has non string standard name, Attempting cast to string",
            stacklevel=2,
        )
        try:
            standard_name = str(standard_name_test)
        except ValueError:
            warnings.warn(
                "Unable to cast standard name to string, excluding from geophysical variables",
                stacklevel=2,
            )
    else:
        standard_name = standard_name_test

    # Is the standard name associated with coordinates
    if standard_name in {
        "time",
        "latitude",
        "longitude",
        "height",
        "depth",
        "altitude",
    }:
        return False

    if variable in get_coordinate_variables(nc):
        return False

    if variable in get_auxiliary_coordinate_variables(nc):
        return False

    if variable in get_forecast_metadata_variables(nc):
        return False

    # Is it dimensionless and unitless?
    if len(ncvar.shape) == 0 and unitless:
        return False

    # Is it a QC Flag?
    if "status_flag" in standard_name or hasattr(ncvar, "flag_meanings"):
        return False

    # Is it a §7.1 Cell Boundaries variable
    if variable in get_cell_boundary_variables(nc):
        return False

    if variable == get_climatology_variable(nc):
        return False

    # Is it a string but with no defined units?
    if hasattr(ncvar.dtype, "char") and ncvar.dtype.char == "S":
        return False
    elif ncvar.dtype is str:
        return False

    # Is it an instrument descriptor?
    if variable in get_instrument_variables(nc):
        return False

    # What about a platform descriptor?
    if variable in get_platform_variables(nc):
        return False

    # Skip count/index variables too
    if hasattr(ncvar, "sample_dimension") or hasattr(ncvar, "instance_dimension"):
        return False

    return True


@lru_cache
def get_coordinate_variables(nc):
    """
    Returns a list of variable names that identify as coordinate variables.

    A coordinate variable is a netCDF variable with exactly one dimension. The
    name of this dimension must be equivalent to the variable name.

    From CF §1.2 Terminology

    It is a one-dimensional variable with the same name as its dimension [e.g.,
    time(time) ], and it is defined as a numeric data type with values that are
    ordered monotonically. Missing values are not allowed in coordinate
    variables.

    :param netCDF4.Dataset nc: An open netCDF dataset
    """
    coord_vars = []
    for dimension in nc.dimensions:
        if dimension in nc.variables:
            # TODO: Handle string coordinate variables
            if nc.variables[dimension].dimensions == (dimension,):
                coord_vars.append(dimension)
    return coord_vars


def get_auxiliary_coordinate_variables(nc):
    """
    Returns a list of auxiliary coordinate variables

    An auxiliary coordinate variable is any netCDF variable that contains
    coordinate data, but is not a coordinate variable (in the sense of the term
    defined by CF).

    :param netCDf4.Dataset nc: An open netCDF dataset
    """

    aux_vars = []
    # get any variables referenced by the coordinates attribute
    for ncvar in nc.get_variables_by_attributes(
        coordinates=lambda x: isinstance(x, str),
    ):
        # split the coordinates into individual variable names
        referenced_variables = ncvar.coordinates.split(" ")
        # if the variable names exist, add them
        for referenced_variable in referenced_variables:
            if (
                referenced_variable in nc.variables
                and referenced_variable not in aux_vars
            ):
                aux_vars.append(referenced_variable)
    # axis variables are automatically in
    for variable in get_axis_variables(nc):
        if variable not in aux_vars:
            aux_vars.append(variable)
    # Last are any variables that define the common coordinate standard names
    coordinate_standard_names = [
        "time",
        "longitude",
        "latitude",
        "height",
        "depth",
        "altitude",
    ]
    coordinate_standard_names += DIMENSIONLESS_VERTICAL_COORDINATES
    # Some datasets like ROMS use multiple variables to define coordinates
    for ncvar in nc.get_variables_by_attributes(
        standard_name=lambda x: x in coordinate_standard_names,
    ):
        if ncvar.name not in aux_vars:
            aux_vars.append(ncvar.name)
    # Remove any that are purely coordinate variables
    ret_val = []
    for aux_var in aux_vars:
        if nc.variables[aux_var].dimensions == (aux_var,):
            continue
        ret_val.append(aux_var)

    return ret_val


def get_forecast_metadata_variables(nc):
    """
    Returns a list of variables that represent forecast reference time
    metadata.

    :param netCDF4.Dataset nc: An open netCDF4 Dataset.
    :rtype: list
    """
    forecast_metadata_standard_names = {
        "forecast_period",
        "forecast_reference_time",
    }
    forecast_metadata_variables = []
    for varname in nc.variables:
        standard_name = getattr(nc.variables[varname], "standard_name", None)
        if standard_name in forecast_metadata_standard_names:
            forecast_metadata_variables.append(varname)
    return forecast_metadata_variables


def get_cell_boundary_map(nc):
    """
    Returns a dictionary mapping a variable to its boundary variable. The
    returned dictionary maps a string variable name to the name of the boundary
    variable.

    :param netCDF4.Dataset nc: netCDF dataset
    """
    boundary_map = {}
    for variable in nc.get_variables_by_attributes(bounds=lambda x: x is not None):
        if variable.bounds in nc.variables:
            boundary_map[variable.name] = variable.bounds
    return boundary_map


def get_cell_boundary_variables(nc):
    """
    Returns a list of variable names for variables that represent cell
    boundaries through the `bounds` attribute

    :param netCDF4.Dataset nc: netCDF dataset
    """
    boundary_variables = []
    has_bounds = nc.get_variables_by_attributes(bounds=lambda x: x is not None)
    for var in has_bounds:
        if var.bounds in nc.variables:
            boundary_variables.append(var.bounds)
    return boundary_variables


def get_bounds_variables(nc):
    contains_bounds = nc.get_variables_by_attributes(bounds=lambda s: s in nc.variables)
    return {nc.variables[parent_var.bounds] for parent_var in contains_bounds}


def get_geophysical_variables(nc):
    """
    Returns a list of variable names for the variables detected as geophysical
    variables.

    :param netCDF4.Dataset nc: An open netCDF dataset
    """
    parameters = []
    for variable in nc.variables:
        if is_geophysical(nc, variable) and variable not in get_bounds_variables(nc):
            parameters.append(variable)
    return parameters


def get_z_variable(nc):
    """
    Returns the name of the variable that defines the Z axis or height/depth

    :param netCDF4.Dataset nc: netCDF dataset
    """
    z_variables = get_z_variables(nc)
    if not z_variables:
        return None

    # Priority is standard_name, units
    for var in z_variables:
        ncvar = nc.variables[var]
        if getattr(ncvar, "standard_name", None) in ("depth", "height", "altitude"):
            return var

    for var in z_variables:
        ncvar = nc.variables[var]
        units = getattr(ncvar, "units", None)
        if isinstance(units, str):
            if units_convertible(units, "bar"):
                return var
            if units_convertible(units, "m"):
                return var

    return z_variables[0]


def get_z_variables(nc):
    """
    Returns a list of all variables matching definitions for Z

    :param netcdf4.dataset nc: an open netcdf dataset object
    """
    z_variables = []
    # Vertical coordinates will be identifiable by units of pressure or the
    # presence of the positive attribute with a value of up/down

    # optionally, the vertical type may be indicated by providing the
    # standard_name attribute or axis='Z'

    total_coords = get_coordinate_variables(nc) + get_auxiliary_coordinate_variables(nc)
    for coord_name in total_coords:
        if coord_name in z_variables:
            continue
        coord_var = nc.variables[coord_name]
        units = getattr(coord_var, "units", None)
        positive = getattr(coord_var, "positive", None)
        standard_name = getattr(coord_var, "standard_name", None)
        axis = getattr(coord_var, "axis", None)
        # If there are no units, we can't identify it as a vertical coordinate
        # by checking pressure or positive
        if units is not None:
            if units_convertible(units, "bar"):
                z_variables.append(coord_name)
            elif isinstance(positive, str):
                if positive.lower() in ["up", "down"]:
                    z_variables.append(coord_name)
        # if axis='Z' we're good
        if coord_name not in z_variables and axis == "Z":
            z_variables.append(coord_name)
        if coord_name not in z_variables and standard_name in (
            "depth",
            "height",
            "altitude",
        ):
            z_variables.append(coord_name)
        if (
            coord_name not in z_variables
            and standard_name in DIMENSIONLESS_VERTICAL_COORDINATES
        ):
            z_variables.append(coord_name)

    return z_variables


def get_lat_variable(nc):
    """
    Returns the first variable matching latitude

    :param netcdf4.dataset nc: an open netcdf dataset object
    """
    latitudes = get_latitude_variables(nc)
    if latitudes:
        return latitudes[0]
    return None


def get_latitude_variables(nc):
    """
    Returns a list of all variables matching definitions for latitude

    :param netcdf4.dataset nc: an open netcdf dataset object
    """
    latitude_variables = []
    # standard_name takes precedence
    for variable in nc.get_variables_by_attributes(standard_name="latitude"):
        latitude_variables.append(variable.name)

    # Then axis
    for variable in nc.get_variables_by_attributes(axis="Y"):
        if not (
            variable.name in latitude_variables
            or getattr(variable, "standard_name", None)
            in {"projection_y_coordinate", "projection_y_angular_coordinate"}
        ):
            latitude_variables.append(variable.name)

    check_fn = partial(
        attr_membership,
        value_set=VALID_LAT_UNITS,
        modifier_fn=lambda s: s.lower(),
    )
    for variable in nc.get_variables_by_attributes(units=check_fn):
        if variable.name not in latitude_variables:
            latitude_variables.append(variable.name)

    return latitude_variables


def get_true_latitude_variables(nc):
    """
    Returns a list of variables defining true latitude.

    CF Chapter 4 refers to latitude as a coordinate variable that can also be
    used in non-standard coordinate systems like rotated pole and other
    projections. Chapter 5 refers to a concept of true latitude where the
    variable defines latitude in a standard projection.

    True latitude, for lack of a better definition, is simply latitude where
    the standard_name is latitude or the units are degrees_north.

    :param netCDF4.Dataset nc: An open netCDF dataset
    """
    lats = get_latitude_variables(nc)
    true_lats = []
    for lat in lats:
        standard_name = getattr(nc.variables[lat], "standard_name", None)
        units = getattr(nc.variables[lat], "units", None)
        if standard_name == "latitude":
            true_lats.append(lat)
        elif isinstance(units, str) and units.lower() in VALID_LAT_UNITS:
            true_lats.append(lat)
    return true_lats


def get_lon_variable(nc):
    """
    Returns the variable for longitude

    :param netCDF4.Dataset nc: netCDF dataset
    """
    longitudes = get_longitude_variables(nc)
    if longitudes:
        return longitudes[0]
    return None


def get_longitude_variables(nc):
    """
    Returns a list of all variables matching definitions for longitude

    :param netcdf4.dataset nc: an open netcdf dataset object
    """
    longitude_variables = []
    # standard_name takes precedence
    for variable in nc.get_variables_by_attributes(standard_name="longitude"):
        longitude_variables.append(variable.name)

    # Then axis
    for variable in nc.get_variables_by_attributes(axis="X"):
        if not (
            variable.name in longitude_variables
            or getattr(variable, "standard_name", None)
            in {"projection_x_coordinate", "projection_x_angular_coordinate"}
        ):
            longitude_variables.append(variable.name)

    check_fn = partial(
        attr_membership,
        value_set=VALID_LON_UNITS,
        modifier_fn=lambda s: s.lower(),
    )
    for variable in nc.get_variables_by_attributes(units=check_fn):
        if variable.name not in longitude_variables:
            longitude_variables.append(variable.name)

    return longitude_variables


def get_true_longitude_variables(nc):
    """
    Returns a list of variables defining true longitude.

    CF Chapter 4 refers to longitude as a coordinate variable that can also be
    used in non-standard coordinate systems like rotated pole and other
    projections. Chapter 5 refers to a concept of true longitude where the
    variable defines longitude in a standard projection.

    True longitude, for lack of a better definition, is simply longitude where
    the standard_name is longitude or the units are degrees_north.

    :param netCDF4.Dataset nc: An open netCDF dataset
    """
    lons = get_longitude_variables(nc)
    true_lons = []
    for lon in lons:
        standard_name = getattr(nc.variables[lon], "standard_name", None)
        units = getattr(nc.variables[lon], "units", None)
        if standard_name == "longitude":
            true_lons.append(lon)
        elif isinstance(units, str) and units.lower() in VALID_LON_UNITS:
            true_lons.append(lon)
    return true_lons


def get_platform_variables(nc):
    """
    Returns a list of platform variable NAMES

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    candidates = []
    for variable in nc.variables:
        platform = getattr(nc.variables[variable], "platform", "")
        if platform and platform in nc.variables:
            if platform not in candidates:
                candidates.append(platform)

    platform = getattr(nc, "platform", "")
    if platform and platform in nc.variables:
        if platform not in candidates:
            candidates.append(platform)
    return candidates


def get_instrument_variables(nc):
    """
    Returns a list of instrument variables

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    candidates = []
    for variable in nc.variables:
        instrument = getattr(nc.variables[variable], "instrument", "")
        if instrument and instrument in nc.variables:
            if instrument not in candidates:
                candidates.append(instrument)

    instrument = getattr(nc, "instrument", "")
    if instrument and instrument in nc.variables:
        if instrument not in candidates:
            candidates.append(instrument)
    return candidates


def get_time_variable(nc):
    """
    Returns the likeliest variable to be the time coordinate variable

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    for var in nc.variables:
        if getattr(nc.variables[var], "axis", "") == "T":
            return var
    else:
        candidates = nc.get_variables_by_attributes(standard_name="time")
        if len(candidates) == 1:
            return candidates[0].name
        else:  # Look for a coordinate variable time
            for candidate in candidates:
                if candidate.dimensions == (candidate.name,):
                    return candidate.name

    # If we still haven't found the candidate
    time_variables = set(get_time_variables(nc))
    coordinate_variables = set(get_coordinate_variables(nc))
    if len(time_variables.intersection(coordinate_variables)) == 1:
        return list(time_variables.intersection(coordinate_variables))[0]

    auxiliary_coordinates = set(get_auxiliary_coordinate_variables(nc))
    if len(time_variables.intersection(auxiliary_coordinates)) == 1:
        return list(time_variables.intersection(auxiliary_coordinates))[0]
    return None


def get_time_variables(nc):
    """
    Returns a list of variables describing the time coordinate

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    time_variables = set()
    for variable in nc.get_variables_by_attributes(standard_name="time"):
        time_variables.add(variable.name)

    for variable in nc.get_variables_by_attributes(axis="T"):
        if variable.name not in time_variables:
            time_variables.add(variable.name)

    regx = r"^(?:day|d|hour|hr|h|minute|min|second|s)s? since .*$"
    for variable in nc.get_variables_by_attributes(units=lambda x: isinstance(x, str)):
        if re.match(regx, variable.units) and variable.name not in time_variables:
            time_variables.add(variable.name)

    return time_variables


def get_axis_variables(nc):
    """
    Returns a list of variables that define an axis of the dataset

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    axis_variables = []
    for ncvar in nc.get_variables_by_attributes(axis=lambda x: x is not None):
        axis_variables.append(ncvar.name)
    return axis_variables


def get_climatology_variable(nc):
    """
    Returns the variable describing climatology bounds if it exists.

    Climatology variables are similar to cell boundary variables that describe
    the climatology bounnc.

    See Example 7.8 in CF 1.6

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    :rtype: str or None
    """
    time = get_time_variable(nc)
    # If there's no time dimension there's no climatology bounds
    if not time:
        return None
    # Climatology variable is simply whatever time points to under the
    # `climatology` attribute.
    if hasattr(nc.variables[time], "climatology"):
        if nc.variables[time].climatology in nc.variables:
            return nc.variables[time].climatology
    return None


def _find_standard_name_modifier_variables(nc, return_deprecated=False):
    def match_modifier_variables(standard_name_str):
        if standard_name_str is None:
            return False
        if not return_deprecated:
            matches = re.search(r"^\w+ +\w+", standard_name_str)
        else:
            matches = re.search(
                r"^\w+ +(?:status_flag|number_of_observations)$",
                standard_name_str,
            )
        return bool(matches)

    return [
        var.name
        for var in nc.get_variables_by_attributes(
            standard_name=match_modifier_variables,
        )
    ]


def get_flag_variables(nc):
    """
    Returns a list of variables that are defined as flag variables

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    flag_variables = []
    for name, ncvar in nc.variables.items():
        standard_name = getattr(ncvar, "standard_name", None)
        if isinstance(standard_name, str) and "status_flag" in standard_name:
            flag_variables.append(name)
        elif hasattr(ncvar, "flag_meanings"):
            flag_variables.append(name)
    return flag_variables


def get_grid_mapping_variables(nc):
    """
    Returns a list of grid mapping variables

    :param netCDF4.Dataset nc: An open netCDF4 Dataset
    """
    grid_mapping_variables = set()
    for ncvar in nc.get_variables_by_attributes(grid_mapping=lambda x: x is not None):
        if ncvar.grid_mapping in nc.variables:
            grid_mapping_variables.add(ncvar.grid_mapping)
    return grid_mapping_variables


def get_axis_map(nc, variable):
    """
    Returns an axis_map dictionary that contains an axis key and the coordinate
    names as values.

    For example::

        {'X': ['longitude'], 'Y': ['latitude'], 'T': ['time']}

    The axis C is for compressed coordinates like a reduced grid, and U is for
    unknown axis. This can sometimes be physical quantities representing a
    continuous discrete axis, like temperature or density.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: Variable name
    """
    all_coords = get_coordinate_variables(nc) + get_auxiliary_coordinate_variables(nc)

    latitudes = get_latitude_variables(nc)
    longitudes = get_longitude_variables(nc)
    times = get_time_variables(nc)
    heights = get_z_variables(nc)

    coordinates = getattr(nc.variables[variable], "coordinates", None)
    if not isinstance(coordinates, str):
        coordinates = []
    else:
        coordinates = coordinates.split(" ")

    # For example
    # {'x': ['longitude'], 'y': ['latitude'], 't': ['time']}
    axis_map = defaultdict(list)
    for coord_name in all_coords:
        axis = getattr(nc.variables[coord_name], "axis", None)
        if not axis or axis not in ("X", "Y", "Z", "T"):
            if is_compression_coordinate(nc, coord_name):
                axis = "C"
            elif coord_name in times:
                axis = "T"
            elif coord_name in longitudes:
                axis = "X"
            elif coord_name in latitudes:
                axis = "Y"
            elif coord_name in heights:
                axis = "Z"
            else:
                axis = "U"

        if coord_name in nc.variables[variable].dimensions:
            if coord_name not in axis_map[axis]:
                axis_map[axis].append(coord_name)

        elif coord_name in coordinates:
            if coord_name not in axis_map[axis]:
                axis_map[axis].append(coord_name)
    return axis_map


def is_coordinate_variable(nc, variable):
    """
    Returns True if the variable is a coordinate variable

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: Variable name
    """
    if variable not in nc.variables:
        return False
    return nc.variables[variable].dimensions == (variable,)


def is_compression_coordinate(nc, variable):
    """
    Returns True if the variable is a coordinate variable that defines a
    compression scheme.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: Variable name
    """
    # Must be a coordinate variable
    if not is_coordinate_variable(nc, variable):
        return False
    # must have a string attribute compress
    compress = getattr(nc.variables[variable], "compress", None)
    if not isinstance(compress, str):
        return False
    if not compress:
        return False
    # This should never happen or be allowed
    if variable in compress:
        return False
    # Must point to dimensions
    for dim in compress.split():
        if dim not in nc.dimensions:
            return False
    return True


def coordinate_dimension_matrix(nc):
    """
    Returns a dictionary of coordinates mapped to their dimensions

    :param netCDF4.Dataset nc: An open netCDF dataset
    """
    retval = {}
    x = get_lon_variable(nc)
    if x:
        retval["x"] = nc.variables[x].dimensions
    y = get_lat_variable(nc)
    if y:
        retval["y"] = nc.variables[y].dimensions

    z = get_z_variable(nc)
    if z:
        retval["z"] = nc.variables[z].dimensions

    t = get_time_variable(nc)
    if t:
        retval["t"] = nc.variables[t].dimensions
    return retval


def is_dataset_valid_ragged_array_repr_featureType(nc, feature_type: str):
    """
    Check if a data set is a valid representation of a ragged
    array structure. See inline comments.
    """

    # regardless of if compound type or not, must have a cf_role
    # variable; if compound, this will be the first part of the
    # feature_type as we'll have to search for one with profile_id
    # regardless; if single feature type, cf_role must match that
    # featureType
    cf_role_vars = nc.get_variables_by_attributes(cf_role=lambda x: x is not None)
    is_compound = False
    if feature_type.lower() in {"timeseriesprofile", "trajectoryprofile"}:
        is_compound = True
        ftype = feature_type.lower().split("profile")[0]
        if len(cf_role_vars) > 2:
            return False
    else:
        ftype = feature_type.lower()
        if len(cf_role_vars) > 1 or not ftype:
            return False

    cf_role_var = nc.get_variables_by_attributes(cf_role=f"{ftype}_id")[0]
    # if cf_role_var returns None, this should raise an error?
    if cf_role_var.cf_role.split("_id")[0].lower() != ftype:
        return False

    # now we'll check dimensions for singular feature types and/or
    # the first half of the compound featureType
    instance_dim = cf_role_var.dimensions
    if len(instance_dim) != 1:
        return False

    # Now we check for the presence of an index variable or count variable;
    # NOTE that if no index or count variables exist, we can't determine with
    # certainty that this is invalid, because single-instance data sets
    # are valid representations of the ragged array structures. Instead,
    # if the index/count variable is present, we check that only one of
    # each is present and that their dimensions are correct
    index_vars = nc.get_variables_by_attributes(
        instance_dimension=lambda x: x is not None,
    )
    count_vars = nc.get_variables_by_attributes(
        sample_dimension=lambda x: x is not None,
    )

    # if the featureType isn't compound, shouldn't have both count and index
    if index_vars and count_vars and not is_compound:
        return False

    # single featureType, checking for valid index variable
    elif index_vars and not is_compound:
        if len(index_vars) > 1:
            return False
        # the index variable's attr 'instance_dimension'
        # must be the same as the actual instance dimension,
        # which we get from the cf_role variable
        if index_vars[0].instance_dimension != instance_dim[0]:
            return False

    # single featureType, checking for valid count variable
    elif count_vars and not is_compound:
        if len(count_vars) > 1:
            return False
        # the count variable must have the same dimensions
        # as the instance variable, which has the instance
        # dimension as its dimension
        if count_vars[0].dimensions != instance_dim:
            return False

    # Now, if the featureType is compound, an index variable
    # must be present for the profile variable. To verify this, we will
    # check that the dimension of the index variable is the same dimension
    # that is present on the variable which has the attribute cf_role=profile_id.
    # The attribute of the index variable 'instance_dimension' should point to the
    # name of the dimension of the cf_role variable for either timeSeries or trajectory.
    # A count variable must also be present, and should have the same dimension,
    # but its attribute 'sample_dimension' must refer to the dimension, which is
    # DIFFERENT than the variable with the attribute cf_role=ftype, where ftype is the
    # first half of the compound featureType (so either timeseries or trajectory).
    # Thus, the dimension of the count variable must be the same dimension as the
    # dimension that all the other geophysical variables have.
    elif index_vars and count_vars and is_compound:
        if len(index_vars) > 1 or len(count_vars) > 1:
            return False

        profile_cf_role_vars = nc.get_variables_by_attributes(cf_role="profile_id")
        if len(profile_cf_role_vars) > 1:
            return False
        profile_cf_role_var = profile_cf_role_vars[0]

        # we first check the dimension of the index variable
        if index_vars[0].dimensions != profile_cf_role_var.dimensions:
            return False

        # the attribute 'instance_dimension' must point to the dimension
        # of the timeseries or trajectory cf_role var
        if index_vars[0].instance_dimension != cf_role_var.dimensions[0]:
            return False

        # get all geophysical dims
        geophysical_dims = [nc[v].dimensions for v in get_geophysical_variables(nc)]
        if len(set(geophysical_dims)) != 1:
            return False

        # check the dimension of the count var is the same and that the
        # sample_dimension attribute points to the same dimension that
        # the geophysical variables have
        if (
            count_vars[0].dimensions != profile_cf_role_var.dimensions
            or (count_vars[0].sample_dimension,) != geophysical_dims[0]
        ):
            return False

    else:
        return False

    return True


def resolve_ragged_array_dimension(ds: Dataset):
    # TODO: put in loop?
    ragged_variable = ds.get_variables_by_attributes(
        sample_dimension=lambda s: isinstance(s, str),
    )
    if ragged_variable:
        ragged_type = "sample_dimension"
    else:
        ragged_variable = ds.get_variables_by_attributes(
            instance_dimension=lambda s: isinstance(s, str),
        )
        ragged_type = "instance_dimension"
    if ragged_variable is None:
        raise ValueError("Could not find a ragged array related variable")
    return ragged_type


def is_variable_valid_ragged_array_repr_featureType(nc, variable: str) -> bool:
    """
    This method returns a boolean indicating whether the variable
    is a valid member of a contiguous ragged array representation
    or indexed ragged array representation of any of the three singular
    CF featureType types.

    For any ragged array representation, any DATA VARIABLE must have
    the sample dimension as its sole dimension. Additionally, for any
    featureType or compound featureType (e.g. timeSeriesProfile), the
    data variable must have the sample dimension as its dimension.
    """

    # Get all geophysical variables; should have only one
    # dimension in the set, and the dimension of the variable
    # should be equal.
    geo_vars = get_geophysical_variables(nc)
    dim_tuples = [nc.variables[v].dimensions for v in geo_vars]
    if len(set(dim_tuples)) < 1:
        return False

    # NOTE
    # Each dimension tuple - there should be only one - should only
    # have the sample dimension as its sole value. If there are more
    # than one, we assume the sample dimension is the first. Is this
    # an appropriate assumption?
    dim = dim_tuples[0]
    if len(dim) != 1:
        return False

    # this is the only thing we have to work with
    return nc.variables[variable].dimensions == dim


def is_point(nc, variable):
    """
    Returns true if the variable is a point feature type

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(o), y(o), z(o), t(o)
    # X(o)

    dims = nc.variables[variable].dimensions

    cmatrix = coordinate_dimension_matrix(nc)
    first_coord = None
    if "t" in cmatrix:
        first_coord = cmatrix["t"]
        if len(cmatrix["t"]) > 1:
            return False
    if "x" in cmatrix:
        if first_coord is None:
            first_coord = cmatrix["x"]
        if first_coord != cmatrix["x"]:
            return False
        if len(cmatrix["x"]) > 1:
            return False
    if "y" in cmatrix:
        if first_coord is None:
            first_coord = cmatrix["y"]
        if first_coord != cmatrix["y"]:
            return False
        if len(cmatrix["y"]) > 1:
            return False
    if "z" in cmatrix:
        if first_coord is None:
            first_coord = cmatrix["z"]
        if first_coord != cmatrix["z"]:
            return False
        if len(cmatrix["z"]) > 1:
            return False
    if first_coord and dims != first_coord:
        return False
    # Point is indistinguishable from trajectories where the instance dimension
    # is implied (scalar)
    traj_ids = nc.get_variables_by_attributes(cf_role="trajectory_id")
    if traj_ids:
        return False

    return True


def is_timeseries(nc, variable):
    """
    Returns true if the variable is a time series feature type.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """

    # x, y, z, t(t)
    # X(t)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)
    time_variables = get_time_variables(nc)

    if len(dims) != 1:
        return False
    dim = dims[0]
    if dim not in time_variables:
        return False
    # No other coordinates can vary with time
    if "x" in cmatrix:
        if len(cmatrix["x"]) != 0:
            return False
    if "y" in cmatrix:
        if len(cmatrix["y"]) != 0:
            return False
    if "z" in cmatrix:
        if len(cmatrix["z"]) != 0:
            return False

    return True


def is_multi_timeseries_orthogonal(nc, variable):
    """
    Returns true if the variable is a orthogonal multidimensional array
    representation of time series. For more information on what this means see
    CF 1.6 §H.2.1

    http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#_orthogonal_multidimensional_array_representation_of_time_series

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i), y(i), z(i), t(o)
    # X(i, o)
    dims = nc.variables[variable].dimensions

    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 1 or cmatrix["x"] != cmatrix["y"]:
        return False
    if "z" in cmatrix and cmatrix["x"] != cmatrix["z"]:
        return False

    timevar = get_time_variable(nc)
    if cmatrix["t"] != (timevar,):
        return False

    i = cmatrix["x"][0]
    o = cmatrix["t"][0]
    if dims == (i, o):
        return True
    return False


def is_multi_timeseries_incomplete(nc, variable):
    """
    Returns true if the variable is an incomplete multidimensional array
    representation of time series. For more information on what this means see
    CF 1.6 §H.2.2

    http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#_incomplete_multidimensional_array_representation_of_time_series

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """

    # x(i), y(i), z(i), t(i, o)
    # X(i, o)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if len(cmatrix["t"]) != 2:
        return False
    if cmatrix["x"][0] != cmatrix["t"][0]:
        return False

    i = cmatrix["x"][0]
    o = cmatrix["t"][1]

    if dims == (i, o):
        return True
    return False


def isTimeSeries(nc, variable):
    """
    Attempt to consolidate all of the disparate timeseries checks.
    Is this being pragmatic or lazy? I'll argue pragmatic.
    Confidently pragmatic.
    """

    # first three check if variable is a valid multidimensional
    # representation, the last checks if it's a valid ragged array
    if (
        is_timeseries(nc, variable)
        or is_multi_timeseries_orthogonal(nc, variable)
        or is_multi_timeseries_incomplete(nc, variable)
    ):
        return True

    return False


def is_cf_trajectory(nc, variable):
    """
    Returns true if the variable is a CF trajectory feature type

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i, o), y(i, o), z(i, o), t(i, o)
    # X(i, o)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 2:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if cmatrix["x"] != cmatrix["t"]:
        return False
    if "z" in cmatrix and cmatrix["x"] != cmatrix["z"]:
        return False
    if dims == cmatrix["x"]:
        return True
    return False


def is_single_trajectory(nc, variable):
    """
    Returns true if the variable is a single trajectory feature

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(o), y(o), z(o), t(o)
    # X(o)
    # cf_role must be trajectory
    dims = nc.variables[variable].dimensions

    cmatrix = coordinate_dimension_matrix(nc)
    # Time is required for trajectories
    if "t" not in cmatrix:
        return False
    # Coordinates changing with time are optional
    if "x" in cmatrix:
        if cmatrix["x"] != cmatrix["t"]:
            return False
    if "y" in cmatrix:
        if cmatrix["y"] != cmatrix["t"]:
            return False
    if "z" in cmatrix:
        if cmatrix["z"] != cmatrix["t"]:
            return False
    if dims != cmatrix["t"]:
        return False
    traj_ids = nc.get_variables_by_attributes(cf_role="trajectory_id")
    if len(traj_ids) != 1:
        return False
    return True


def isTrajectory(nc, variable):
    """
    Wrapper method for checking if a variable is detected as
    a trajectory featureType.
    """

    if is_cf_trajectory(nc, variable) or is_single_trajectory(nc, variable):
        return True

    return False


def is_profile_orthogonal(nc, variable):
    """
    Returns true if the variable is a orthogonal profile feature type

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # Every profile has the exact same depths, think thermister or ADCP
    # x(i), y(i), z(j), t(i)
    # X(i, j)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if cmatrix["x"] != cmatrix["t"]:
        return False
    if len(cmatrix["z"]) != 1:
        return False

    i = cmatrix["x"][0]
    j = cmatrix["z"][0]

    if dims == (i, j):
        return True
    return False


def is_profile_incomplete(nc, variable):
    """
    Returns true if the variable is a incomplete profile feature type

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # Every profile may have different depths
    # x(i), y(i), z(i, j), t(i)
    # X(i, j)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if cmatrix["x"] != cmatrix["t"]:
        return False
    if len(cmatrix["z"]) != 2:
        return False
    if cmatrix["z"][0] != cmatrix["x"][0]:
        return False

    i = cmatrix["x"][0]
    j = cmatrix["z"][1]

    if dims == (i, j):
        return True
    return False


def isProfile(nc, variable: str):
    """
    Per Ch 9 of the CF spec, profiles are a part of the CF Discrete
    Sampling Geometries. Profile data can be logically represented
    in a file one of four ways: orthogonal multidimensional array,
    incomplete multidimensional array, contiguous ragged array,
    and indexed ragged array. If the variable is found to be any
    valid for any one of these representations, return "profile"
    else return None.

    The very first part of this function attempts to use legacy code
    to test if the variable is a profile in the orthogonal multidimensional
    array representation or the incomplete multidimensional array
    representation. If neither of these are true, it moves on to testing
    for the contiguous ragged array and indexed ragged array representations.

    Parameters
    ----------
    nc      : netCDF4 Dataset
    variable: str name of variable

    Returns
    -------
    str or None
    """

    # NOTE
    # TODO
    # Does this take into account a single profile? This is a valid profile.

    # first check for orthogonal, incomplete
    if is_profile_orthogonal(nc, variable) or is_profile_incomplete(nc, variable):
        return True


def is_timeseries_profile_single_station(nc, variable):
    """
    Returns true if the variable is a time-series profile that represents a
    single station and each profile is the same length.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """

    # x, y, z(z), t(t)
    # X(t, z)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 0:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False

    z = get_z_variable(nc)
    if cmatrix["z"] != (z,):
        return False
    t = get_time_variable(nc)
    if cmatrix["t"] != (t,):
        return False

    if dims == (t, z):
        return True
    return False


def is_timeseries_profile_multi_station(nc, variable):
    """
    Returns true if the variable is a time-series profile that represents multiple stations with orthogonal time and depth

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i), y(i), z(z), t(t)
    # X(i, t, z)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False
    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    i = cmatrix["x"][0]

    z = get_z_variable(nc)
    if cmatrix["z"] != (z,):
        return False
    t = get_time_variable(nc)
    if cmatrix["t"] != (t,):
        return False

    if dims == (i, t, z):
        return True
    return False


def is_timeseries_profile_single_ortho_time(nc, variable):
    """
    Returns true if the variable is a time-series profile that represents a
    single station with orthogonal time only.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x, y, z(t, j), t(t)
    # X(t, j)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 0:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False

    t = get_time_variable(nc)
    if cmatrix["t"] != (t,):
        return False

    if len(cmatrix["z"]) != 2:
        return False

    if cmatrix["z"][0] != t:
        return False

    j = cmatrix["z"][1]

    if dims == (t, j):
        return True
    return False


def is_timeseries_profile_multi_ortho_time(nc, variable):
    """
    Returns true if the variable is a time-series profile that represents a
    multi station with orthogonal time only.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i), y(i), z(i, t, j), t(t)
    # X(i, t, j)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False

    t = get_time_variable(nc)
    if cmatrix["t"] != (t,):
        return False

    if len(cmatrix["z"]) != 3:
        return False

    if cmatrix["z"][1] != t:
        return False
    if cmatrix["z"][0] != cmatrix["x"][0]:
        return False

    i = cmatrix["x"][0]
    j = cmatrix["z"][2]

    if dims == (i, t, j):
        return True
    return False


def is_timeseries_profile_ortho_depth(nc, variable):
    """
    Returns true if the variable is a time-series profile with orthogonal depth
    only.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i), y(i), z(z), t(i, j)
    # X(i, j, z)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False

    z = get_z_variable(nc)
    if cmatrix["z"] != (z,):
        return False

    i = cmatrix["x"][0]

    if len(cmatrix["t"]) != 2:
        return False
    if cmatrix["t"][0] != i:
        return False

    j = cmatrix["t"][1]

    if dims == (i, j, z):
        return True
    return False


def is_timeseries_profile_incomplete(nc, variable):
    """
    Returns true if the variable is a time-series profile incomplete depth and
    incomplete time.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i), y(i), z(i, j, k), t(i, j)
    # X(i, j, k)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 1:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    i = cmatrix["x"][0]

    if len(cmatrix["t"]) != 2:
        return False
    if cmatrix["t"][0] != i:
        return False
    j = cmatrix["t"][1]

    if len(cmatrix["z"]) != 3:
        return False
    if cmatrix["z"][0] != i:
        return False
    if cmatrix["z"][1] != j:
        return False
    k = cmatrix["z"][2]

    if dims == (i, j, k):
        return True
    return False


def isTimeSeriesProfile(nc, variable):
    """
    Wrapper method.
    Verify if a variable matches with the timeSeriesProfile
    featureType. According to the CF specification, a data set
    with timeSeriesProfile features has two cf_role variables:
    one for the "station" (cf_role=timeseries_id) and one for the
    "profile" (cf_role=profile_id).
    """

    if (
        is_timeseries_profile_single_station(nc, variable)
        or is_timeseries_profile_multi_station(nc, variable)
        or is_timeseries_profile_single_ortho_time(nc, variable)
        or is_timeseries_profile_multi_ortho_time(nc, variable)
        or is_timeseries_profile_ortho_depth(nc, variable)
        or is_timeseries_profile_incomplete(nc, variable)
    ):
        return True

    return False


def is_trajectory_profile_orthogonal(nc, variable):
    """
    Returns true if the variable is a trajectory profile with orthogonal
    depths.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i, o), y(i, o), z(z), t(i, o)
    # X(i, o, z)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 2:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if cmatrix["x"] != cmatrix["t"]:
        return False

    i, o = cmatrix["x"]

    z = get_z_variable(nc)
    if cmatrix["z"] != (z,):
        return False

    if dims == (i, o, z):
        return True
    return False


def is_trajectory_profile_incomplete(nc, variable):
    """
    Returns true if the variable is a trajectory profile with incomplete
    depths.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(i, o), y(i, o), z(i, o, j), t(i, o)
    # X(i, o, j)
    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    if len(cmatrix["x"]) != 2:
        return False
    if cmatrix["x"] != cmatrix["y"]:
        return False
    if cmatrix["x"] != cmatrix["t"]:
        return False

    i, o = cmatrix["x"]

    if len(cmatrix["z"]) != 3:
        return False

    if cmatrix["z"][0] != i:
        return False
    if cmatrix["z"][1] != o:
        return False

    j = cmatrix["z"][2]

    if dims == (i, o, j):
        return True
    return False


def isTrajectoryProfile(nc, variable):
    """
    Wrapper method
    """

    # NOTE
    # does this take into account single trajectory profile?
    if is_trajectory_profile_orthogonal(
        nc,
        variable,
    ) or is_trajectory_profile_incomplete(nc, variable):
        return True

    return False


def is_2d_regular_grid(nc, variable):
    """
    Returns True if the variable is a 2D Regular grid.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(x), y(y), t(t)
    # X(t, y, x)

    if is_mapped_grid(nc, variable):
        return False

    dims = nc.variables[variable].dimensions

    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "t"):
        if req not in cmatrix:
            return False

    x = get_lon_variable(nc)
    y = get_lat_variable(nc)
    t = get_time_variable(nc)

    if cmatrix["x"] != (x,):
        return False
    if cmatrix["y"] != (y,):
        return False
    if cmatrix["t"] != (t,):
        return False

    # Relaxed dimension ordering
    if len(dims) == 3 and x in dims and y in dims and t in dims:
        return True
    return False


def is_2d_static_grid(nc, variable):
    """
    Returns True if the variable is a 2D Regular grid that does not vary with
    time.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(x), y(y)
    # X(y, x)

    if is_mapped_grid(nc, variable):
        return False

    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y"):
        if req not in cmatrix:
            return False

    x = get_lon_variable(nc)
    y = get_lat_variable(nc)

    if cmatrix["x"] != (x,):
        return False

    if cmatrix["y"] != (y,):
        return False

    if len(dims) != 2 or x not in dims or y not in dims:
        return False

    return True


def is_3d_regular_grid(nc, variable):
    """
    Returns True if the variable is a 3D Regular grid.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(x), y(y), z(z), t(t)
    # X(t, z, y, x)

    if is_mapped_grid(nc, variable):
        return False

    dims = nc.variables[variable].dimensions

    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z", "t"):
        if req not in cmatrix:
            return False

    x = get_lon_variable(nc)
    y = get_lat_variable(nc)
    z = get_z_variable(nc)
    t = get_time_variable(nc)

    if cmatrix["x"] != (x,):
        return False
    if cmatrix["y"] != (y,):
        return False
    if cmatrix["z"] != (z,):
        return False
    if cmatrix["t"] != (t,):
        return False

    # Relaxed dimension ordering
    if len(dims) == 4 and x in dims and y in dims and t in dims and z in dims:
        return True
    return False


def is_3d_static_grid(nc, variable):
    """
    Returns True if the variable is a 2D Regular grid that does not vary with
    time.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(x), y(y), z(z)
    # X(z, y, x)

    if is_mapped_grid(nc, variable):
        return False

    dims = nc.variables[variable].dimensions
    cmatrix = coordinate_dimension_matrix(nc)

    for req in ("x", "y", "z"):
        if req not in cmatrix:
            return False

    x = get_lon_variable(nc)
    y = get_lat_variable(nc)
    z = get_z_variable(nc)

    if cmatrix["x"] != (x,):
        return False

    if cmatrix["y"] != (y,):
        return False

    if cmatrix["z"] != (z,):
        return False

    if len(dims) != 3 or x not in dims or y not in dims or z not in dims:
        return False

    return True


def is_mapped_grid(nc, variable):
    """
    Returns true if the feature-type of variable corresponds to a mapped grid
    type. Characterized by Appedix F of CF-1.6

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    # x(j, i), y(j, i), z?, t?
    # F(t?, z?, j, i)

    # The important and really defining characteristic of mapped grids is that
    # the true latitude and longitude coordinates are functions of (j,i) and
    # that the geophysical variables are also functions of (j,i) in their
    # dimensions.
    dims = nc.variables[variable].dimensions
    # For cases like ROMS, the coordinates are mapped using the coordinates attribute
    variable_coordinates = getattr(nc.variables[variable], "coordinates", "").split()

    lons = get_longitude_variables(nc)
    for lon in lons:
        if lon in variable_coordinates:
            break
    else:
        lon = get_lon_variable(nc)

    if lon is None:
        return False

    lats = get_latitude_variables(nc)
    for lat in lats:
        if lat in variable_coordinates:
            break
    else:
        lat = get_lat_variable(nc)

    if lat is None:
        return False

    x = nc.variables[lon].dimensions
    y = nc.variables[lat].dimensions

    if len(x) != 2:
        return False
    if x != y:
        return False

    comma_dimension = ",".join(dims)
    # Dimensions must be in the same order and the mapping coordinates i and j
    # must be in the same order
    if ",".join(x) not in comma_dimension:
        return False

    return True


def is_reduced_grid(nc, variable):
    """
    Returns True if the feature-type of the variable corresponds to a reduced
    horizontal grid.

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    axis_map = get_axis_map(nc, variable)

    if "X" not in axis_map:
        return False
    if "Y" not in axis_map:
        return False
    if "C" not in axis_map:
        return False

    compressed_coordinates = axis_map["C"]
    if len(compressed_coordinates) > 1:
        return False
    compressed_coordinate = axis_map["C"][0]
    for dim in nc.variables[compressed_coordinate].compress.split():
        if dim not in nc.dimensions:
            return False
    return True


def guess_feature_type(nc, variable):
    """
    Returns a string describing the feature type for this variable

    :param netCDF4.Dataset nc: An open netCDF dataset
    :param str variable: name of the variable to check
    """
    if is_point(nc, variable):
        return "point"
    if isProfile(nc, variable):
        return "profile"
    if isTimeSeries(nc, variable):
        return "timeseries"
    if isTrajectory(nc, variable):
        return "trajectory"
    if isTimeSeriesProfile(nc, variable):
        return "timeseriesprofile"
    if isTrajectoryProfile(nc, variable):
        return "trajectoryprofile"

    # TODO
    # consolidate below into "isGrid" ?
    if is_2d_regular_grid(nc, variable):
        return "2d-regular-grid"
    if is_2d_static_grid(nc, variable):
        return "2d-static-grid"
    if is_3d_regular_grid(nc, variable):
        return "3d-regular-grid"
    if is_3d_static_grid(nc, variable):
        return "3d-static-grid"
    if is_mapped_grid(nc, variable):
        return "mapped-grid"
    if is_reduced_grid(nc, variable):
        return "reduced-grid"


def units_convertible(units1, units2):
    """
    Return True if a Unit representing the string units1 can be converted
    to a Unit representing the string units2, else False.

    :param str units1: A string representing the units
    :param str units2: A string representing the units
    """
    try:
        u1 = Unit(units1)
        u2 = Unit(units2)
    except (ValueError, NotImplementedError):
        return False
    return u1.is_convertible(u2)


def get_safe(dict_instance, keypath, default=None):
    """
    Returns a value with in a nested dict structure from a dot separated
    path expression such as "system.server.host" or a list of key entries
    @retval Value if found or None
    """
    try:
        obj = dict_instance
        keylist = keypath if isinstance(keypath, list) else keypath.split(".")
        for key in keylist:
            obj = obj[key]
        return obj
    except Exception:
        return default


class VariableReferenceError(Exception):
    """A variable to assign bad variable references to"""

    def __init__(self, name: str, dataset: Dataset = None):
        self.name = name
        self.dataset_path = dataset.filepath() if dataset is not None else None

    def __str__(self):
        return (
            f"Cannot find variable named {self.name} in dataset " f"{self}.dataset_path"
        )


class StandardNameTable:
    class NameEntry:
        def __init__(self, entrynode):
            self.canonical_units = self._get(entrynode, "canonical_units", True)
            self.grib = self._get(entrynode, "grib")
            self.amip = self._get(entrynode, "amip")
            self.description = self._get(entrynode, "description")

        def _get(self, entrynode, attrname, required=False):
            vals = entrynode.xpath(attrname)
            if len(vals) > 1:
                raise Exception(f"Multiple attrs ({attrname}) found")
            elif required and len(vals) == 0:
                raise Exception(f"Required attr ({attrname}) not found")

            return vals[0].text if len(vals) > 0 else None

    def __init__(self, cached_location=None):
        if cached_location:
            with open(cached_location, encoding="utf-8") as fp:
                resource_text = fp.read()
        elif os.environ.get("CF_STANDARD_NAME_TABLE") and os.path.exists(
            os.environ["CF_STANDARD_NAME_TABLE"],
        ):
            with open(
                os.environ["CF_STANDARD_NAME_TABLE"],
                encoding="utf-8",
            ) as fp:
                resource_text = fp.read()
        else:
            resource_text = get_data(
                "compliance_checker",
                "data/cf-standard-name-table.xml",
            )

        parser = etree.XMLParser(remove_blank_text=True)
        self._root = etree.fromstring(resource_text, parser)

        # generate and save a list of all standard names in file
        self._names = [node.get("id") for node in self._root.iter("entry")]
        self._aliases = [node.get("id") for node in self._root.iter("alias")]
        self._version = self._root.xpath("version_number")[0].text

    def __len__(self):
        return len(self._names) + len(self._aliases)

    def __getitem__(self, key):
        if not (key in self._names or key in self._aliases):
            raise KeyError(f"{key} not found in standard name table")

        if key in self._aliases:
            idx = self._aliases.index(key)
            entryids = self._root.xpath("alias")[idx].xpath("entry_id")

            if len(entryids) != 1:
                raise Exception(
                    f"Inconsistency in standard name table, could not lookup alias for {key}",
                )

            key = entryids[0].text

        if key not in self._names:
            raise KeyError(f"{key} not found in standard name table")

        idx = self._names.index(key)
        entry = self.NameEntry(self._root.xpath("entry")[idx])
        return entry

    def get(self, key, default=None):
        """
        Returns the item for the key or returns the default if it does not exist
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        return key in self._names or key in self._aliases

    def __iter__(self):
        return iter(itertools.chain(self._names, self._aliases))


def download_cf_standard_name_table(version, location=None):
    """
    Downloads the specified CF standard name table version and saves it to file

    :param str version: CF standard name table version number (i.e 34)
    :param str location: Path/filename to write downloaded xml file to
    """

    if (
        location is None
    ):  # This case occurs when updating the packaged version from command line
        location = files("compliance_checker") / "data/cf-standard-name-table.xml"

    if version == "latest":
        url = "http://cfconventions.org/Data/cf-standard-names/current/src/cf-standard-name-table.xml"
    else:
        url = f"http://cfconventions.org/Data/cf-standard-names/{version}/src/cf-standard-name-table.xml"

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()

    print(
        f"Downloading cf-standard-names table version {version} from: {url}",
        file=sys.stderr,
    )
    with open(location, "wb") as f:
        f.write(r.content)


def create_cached_data_dir():
    """
    Returns the path to the data directory to download CF standard names.
    Use $XDG_DATA_HOME.
    """
    writable_directory = os.path.join(os.path.expanduser("~"), ".local", "share")
    data_directory = os.path.join(
        os.environ.get("XDG_DATA_HOME", writable_directory),
        "compliance-checker",
    )
    if not os.path.isdir(data_directory):
        os.makedirs(data_directory)

    return data_directory


def units_known(units):
    try:
        Unit(units)
    except ValueError:
        return False
    return True


def units_temporal(units):
    # Short-circuit for pyudunits2 b/c it is not CF strict as cf-units.
    # Also, this is much faster than creating the units again.
    for noncompliant_preposition in ("@", "after", "from", "ref"):
        if noncompliant_preposition in units:
            return False
    try:
        u = Unit(units)
    except ValueError:
        return False
    # IMPLEMENTATION CONFORMANCE REQUIRED 4.4 1/3
    # time units of a time_coordinate variable must contain a reference
    # time/date
    # IMPLEMENTATION CONFORMANCE REQUIRED 4.4 3/3
    # check that reference time seconds is not greater than or
    # equal to 60
    return u.is_time_reference()


def find_coord_vars(ncds):
    """
    Finds all coordinate variables in a dataset.

    A variable with the same name as a dimension is called a coordinate variable.
    """
    coord_vars = []

    for d in ncds.dimensions:
        if d in ncds.variables and ncds.variables[d].dimensions == (d,):
            coord_vars.append(ncds.variables[d])

    return coord_vars


def is_time_variable(varname, var):
    """
    Identifies if a variable is represents time
    """
    satisfied = varname.lower() == "time"
    satisfied |= getattr(var, "standard_name", "") == "time"
    satisfied |= getattr(var, "axis", "") == "T"
    satisfied |= units_convertible(
        "seconds since 1900-01-01",
        getattr(var, "units", ""),
    )
    return satisfied


def is_vertical_coordinate(var_name, var):
    """
    Determines if a variable is a vertical coordinate variable

    4.3
    A vertical coordinate will be identifiable by: units of pressure; or the presence of the positive attribute with a
    value of up or down (case insensitive).  Optionally, the vertical type may be indicated additionally by providing
    the standard_name attribute with an appropriate value, and/or the axis attribute with the value Z.
    """
    # Known name
    satisfied = var_name.lower() in _possiblez
    satisfied |= getattr(var, "standard_name", "") in _possiblez
    # Is the axis set to Z?
    satisfied |= getattr(var, "axis", "").lower() == "z"
    is_pressure = units_convertible(getattr(var, "units", "1"), "dbar")
    # Pressure defined or positive defined
    satisfied |= is_pressure
    if not is_pressure:
        satisfied |= getattr(var, "positive", "").lower() in ("up", "down")
    return satisfied


def string_from_var_type(variable):
    if isinstance(variable, str):
        return variable[:]
    elif variable.dtype.kind == "S":
        strip_char = variable.fill_value or b"\x00"
        return variable.tobytes().rstrip(strip_char).decode("utf-8")
    else:
        raise TypeError(
            f"Variable '{variable.name} has non-string/character' "
            f"dtype {variable.dtype}",
        )


def get_possible_label_variable_dimensions(variable: Variable) -> tuple[int, ...]:
    """
    Return dimensions if non-char variable, or return variable dimensions
    without trailing dimension if char variable, treating it as a label variable.
    """
    if variable.dtype.kind == "C" and len(variable.dimensions) > 0:
        return variable.dimensions[:-1]
    return variable.dimensions


@lru_cache
def maybe_lateral_reference_variable_or_dimension(
    group: Group | Dataset,
    name: str,
    reference_type: Variable | Dimension,
):

    def can_lateral_search(name):
        return not name.startswith(".") and posixpath.split(name)[0] == ""

    if reference_type == "variable":
        # first try to fetch any
        # can't set to None with .get
        try:
            maybe_var = group[name]
        except IndexError:
            maybe_var = None
        else:
            if isinstance(maybe_var, Variable):
                return maybe_var

        # alphanumeric string by itself, not a relative or absolute
        # search by proximity
        if posixpath.split(name)[0] == "" and not (
            name.startswith(".") or name.startswith("/")
        ):
            group_traverse = group
            while group_traverse.parent:
                group_traverse = group_traverse.parent
                # check_target = posixpath.join(group_traverse.path, name)
                try:
                    maybe_var = group_traverse[name]
                except IndexError:
                    maybe_var = None
                else:
                    if isinstance(maybe_var, Variable):
                        return maybe_var
        else:
            return VariableReferenceError(name)

            # can't find path relative to current group or absolute path
            # perform lateral search if we aren't in the root group


def reference_attr_variables(
    dataset: Dataset,
    attributes_string: str,
    split_by: str = None,
    reference_type: str = "variables",
    group: Group | Dataset = None,
):
    """
    Attempts to reference variables in the string, optionally splitting by
    a string
    """
    references, errors = [], []
    if attributes_string is None:
        return None, None
    if reference_type not in {"variables", "dimensions"}:
        raise ValueError("reference_type must be one of 'variables' or 'dimensions'")
    reference_attr = getattr(dataset, reference_type)
    if split_by is None:
        return_val = reference_attr.get(
            attributes_string,
            VariableReferenceError(attributes_string),
        )
        if not isinstance(return_val, VariableReferenceError):
            return return_val, None
        else:
            return None, return_val
    else:
        string_proc = attributes_string.split(split_by)
        for var_name in string_proc:
            if var_name in reference_attr:
                references.append(reference_attr[var_name])
            else:
                errors.append(VariableReferenceError(var_name))
        return references, errors
